@@ -29,53 +29,66 @@ import (
 	"github.com/apache/skywalking-banyandb/api/common"
 	modelv1 "github.com/apache/skywalking-banyandb/api/proto/banyandb/model/v1"
 	"github.com/apache/skywalking-banyandb/pkg/index"
+	"github.com/apache/skywalking-banyandb/pkg/run"
+	"github.com/apache/skywalking-banyandb/pkg/timestamp"
 )
 
-func (s *store) Sort(sids []common.SeriesID, fieldKey index.FieldKey, order modelv1.Sort, preLoadSize int) (iter index.FieldIterator, err error) {
+func (s *store) Sort(ctx context.Context, sids []common.SeriesID, fieldKey index.FieldKey, order modelv1.Sort,
+	timeRange *timestamp.TimeRange, preLoadSize int,
+) (iter index.FieldIterator[*index.DocumentResult], err error) {
 	reader, err := s.writer.Reader()
 	if err != nil {
 		return nil, err
 	}
-	fk := fieldKey.MarshalIndexRule()
+
 	tqs := make([]bluge.Query, len(sids))
 	for i := range sids {
 		tq := bluge.NewTermQuery(string(sids[i].Marshal()))
 		tq.SetField(seriesIDField)
 		tqs[i] = tq
 	}
+	drq := bluge.
+		NewDateRangeInclusiveQuery(timeRange.Start, timeRange.End, timeRange.IncludeStart, timeRange.IncludeEnd).
+		SetField(timestampField)
 	var query bluge.Query
 	if len(tqs) == 0 {
-		query = bluge.NewMatchAllQuery()
+		query = drq
 	} else {
-		bq := bluge.NewBooleanQuery()
-		bq.AddShould(tqs...)
-		bq.SetMinShould(1)
-		query = bq
+		ibq := bluge.NewBooleanQuery()
+		ibq.AddShould(tqs...)
+		ibq.SetMinShould(1)
+		obq := bluge.NewBooleanQuery()
+		obq.AddMust(ibq)
+		obq.AddMust(drq)
+		query = obq
 	}
 
+	fk := fieldKey.Marshal()
 	sortedKey := fk
 	if order == modelv1.Sort_SORT_DESC {
 		sortedKey = "-" + sortedKey
 	}
 	result := &sortIterator{
-		query:     query,
+		query:     &queryNode{query: query},
 		reader:    reader,
 		sortedKey: sortedKey,
 		size:      preLoadSize,
-		sid:       fieldKey.SeriesID,
+		ctx:       ctx,
 	}
 	return result, nil
 }
 
 type sortIterator struct {
-	query     bluge.Query
+	query     index.Query
 	err       error
+	ctx       context.Context
 	reader    *bluge.Reader
 	current   *blugeMatchIterator
+	closer    *run.Closer
 	sortedKey string
+	fields    []string
 	size      int
 	skipped   int
-	sid       common.SeriesID
 }
 
 func (si *sortIterator) Next() bool {
@@ -99,18 +112,18 @@ func (si *sortIterator) loadCurrent() bool {
 		// overflow
 		size = math.MaxInt64
 	}
-	topNSearch := bluge.NewTopNSearch(size, si.query).SortBy([]string{si.sortedKey})
+	topNSearch := bluge.NewTopNSearch(size, si.query.(*queryNode).query).SortBy([]string{si.sortedKey})
 	if si.skipped > 0 {
 		topNSearch = topNSearch.SetFrom(si.skipped)
 	}
 
-	documentMatchIterator, err := si.reader.Search(context.Background(), topNSearch)
+	documentMatchIterator, err := si.reader.Search(si.ctx, topNSearch)
 	if err != nil {
 		si.err = err
 		return false
 	}
 
-	iter := newBlugeMatchIterator(documentMatchIterator, nil)
+	iter := newBlugeMatchIterator(documentMatchIterator, nil, si.fields)
 	si.current = &iter
 	if si.next() {
 		return true
@@ -127,11 +140,16 @@ func (si *sortIterator) next() bool {
 	return false
 }
 
-func (si *sortIterator) Val() (uint64, common.SeriesID) {
-	return si.current.Val()
+func (si *sortIterator) Val() *index.DocumentResult {
+	v := si.current.Val()
+	if v.SortedValue == nil {
+		panic("sorted field not found in document")
+	}
+	return &v
 }
 
 func (si *sortIterator) Close() error {
+	defer si.closer.Done()
 	if errors.Is(si.err, io.EOF) {
 		si.err = nil
 		if si.current != nil {
@@ -144,3 +162,7 @@ func (si *sortIterator) Close() error {
 	}
 	return errors.Join(si.err, si.current.Close(), si.reader.Close())
 }
+
+func (si *sortIterator) Query() index.Query {
+	return si.query
+}
