@@ -18,9 +18,11 @@
 package inverted
 
 import (
+	"context"
 	"fmt"
 	"sort"
 	"testing"
+	"time"
 
 	"github.com/stretchr/testify/assert"
 	"github.com/stretchr/testify/require"
@@ -32,6 +34,7 @@ import (
 	"github.com/apache/skywalking-banyandb/pkg/index/posting"
 	"github.com/apache/skywalking-banyandb/pkg/index/posting/roaring"
 	"github.com/apache/skywalking-banyandb/pkg/logger"
+	"github.com/apache/skywalking-banyandb/pkg/timestamp"
 )
 
 const indexRuleID = 3
@@ -48,8 +51,8 @@ func TestStore_Sort(t *testing.T) {
 		fn()
 	}()
 	tester.NoError(err)
-	data := setUpDuration(require.New(t), s)
-	s.(*store).flush()
+	now := time.Now()
+	data := setUpDuration(require.New(t), now, s)
 
 	tests := []struct {
 		name string
@@ -148,11 +151,12 @@ func TestStore_Sort(t *testing.T) {
 			})
 		}
 	}
+	tr := timestamp.NewInclusiveTimeRange(now, now)
 	for _, tt := range allTests {
 		t.Run(tt.name, func(t *testing.T) {
 			tester := assert.New(t)
 			is := require.New(t)
-			iter, err := s.Sort(tt.args.sids, index.FieldKey{IndexRuleID: indexRuleID}, tt.args.orderType, tt.preloadSize)
+			iter, err := s.Sort(context.TODO(), tt.args.sids, index.FieldKey{IndexRuleID: indexRuleID}, tt.args.orderType, &tr, tt.preloadSize)
 			is.NoError(err)
 			if iter == nil {
 				tester.Empty(tt.want)
@@ -167,8 +171,11 @@ func TestStore_Sort(t *testing.T) {
 			is.NotNil(iter)
 			var got result
 			for iter.Next() {
-				docID, _ := iter.Val()
-				got.items = append(got.items, docID)
+				val := iter.Val()
+				got.items = append(got.items, val.DocID)
+				if val.SortedValue != nil {
+					got.terms = append(got.terms, val.SortedValue)
+				}
 			}
 			for i := 0; i < 10; i++ {
 				is.False(iter.Next())
@@ -178,7 +185,8 @@ func TestStore_Sort(t *testing.T) {
 				pl := data[w]
 				wants.items = append(wants.items, pl.ToSlice()...)
 			}
-			tester.Equal(wants, got, tt.name)
+			tester.Equal(wants.items, got.items, tt.name)
+			tester.Equal(len(got.items), len(got.terms), tt.name)
 		})
 	}
 }
@@ -190,9 +198,10 @@ type args struct {
 
 type result struct {
 	items []uint64
+	terms [][]byte
 }
 
-func setUpDuration(t *require.Assertions, store index.Writer) map[int]posting.List {
+func setUpDuration(t *require.Assertions, ts time.Time, store index.Writer) map[int]posting.List {
 	r := map[int]posting.List{
 		50:   roaring.NewPostingList(),
 		200:  roaring.NewPostingList(),
@@ -205,6 +214,8 @@ func setUpDuration(t *require.Assertions, store index.Writer) map[int]posting.Li
 		idx = append(idx, key)
 	}
 	sort.Ints(idx)
+
+	var batch index.Batch
 	for i := 100; i < 200; i++ {
 		id := uint64(i)
 		for i2, term := range idx {
@@ -212,15 +223,20 @@ func setUpDuration(t *require.Assertions, store index.Writer) map[int]posting.Li
 				continue
 			}
 			sid := i2%2 + 1
-			t.NoError(store.Write([]index.Field{{
-				Key: index.FieldKey{
-					SeriesID:    common.SeriesID(sid),
-					IndexRuleID: indexRuleID,
-				},
-				Term: convert.Int64ToBytes(int64(term)),
-			}}, id))
+			batch.Documents = append(batch.Documents, index.Document{
+				Fields: []index.Field{{
+					Key: index.FieldKey{
+						SeriesID:    common.SeriesID(sid),
+						IndexRuleID: indexRuleID,
+					},
+					Term: convert.Int64ToBytes(int64(term)),
+				}},
+				DocID:     id,
+				Timestamp: ts.UnixNano(),
+			})
 			r[term].Insert(id)
 		}
 	}
+	t.NoError(store.Batch(batch))
 	return r
 }
