@@ -115,7 +115,6 @@ func parseFromYAML(tryParseGroup bool, reader io.Reader) (requests []reqBody, er
 		}
 
 		var group string
-		// TODO:// bydbctl should support multiple groups
 		if metadata, ok := data["metadata"].(map[string]interface{}); ok {
 			group, ok = metadata["group"].(string)
 			if !ok && tryParseGroup {
@@ -142,7 +141,7 @@ func parseFromYAML(tryParseGroup bool, reader io.Reader) (requests []reqBody, er
 				data["groups"] = []string{group}
 			}
 		} else {
-			return nil, errors.WithMessage(errMalformedInput, "absent node: metadata or name&group")
+			return nil, errors.WithMessage(errMalformedInput, "absent node: name or groups")
 		}
 		j, err = json.Marshal(data)
 		if err != nil {
@@ -309,7 +308,7 @@ func yamlPrinter(index int, _ reqBody, body []byte) error {
 	return nil
 }
 
-func rest(pfn paramsFn, fn reqFn, printer printer, enableTLS bool, insecure bool, grpcCert string) (err error) {
+func rest(pfn paramsFn, fn reqFn, printer printer, enableTLS bool, insecure bool, cert string) (err error) {
 	var requests []reqBody
 	if pfn == nil {
 		requests = []reqBody{{}}
@@ -327,8 +326,8 @@ func rest(pfn paramsFn, fn reqFn, printer printer, enableTLS bool, insecure bool
 				// #nosec G402
 				InsecureSkipVerify: insecure,
 			}
-			if grpcCert != "" {
-				cert, err := os.ReadFile(grpcCert)
+			if cert != "" {
+				cert, err := os.ReadFile(cert)
 				if err != nil {
 					return err
 				}
