@@ -34,19 +34,22 @@ import (
 	"github.com/apache/skywalking-banyandb/banyand/queue"
 	"github.com/apache/skywalking-banyandb/pkg/accesslog"
 	"github.com/apache/skywalking-banyandb/pkg/bus"
+	"github.com/apache/skywalking-banyandb/pkg/convert"
 	"github.com/apache/skywalking-banyandb/pkg/logger"
 	pbv1 "github.com/apache/skywalking-banyandb/pkg/pb/v1"
+	"github.com/apache/skywalking-banyandb/pkg/query"
 	"github.com/apache/skywalking-banyandb/pkg/timestamp"
 )
 
 type streamService struct {
 	streamv1.UnimplementedStreamServiceServer
-	*discoveryService
-	sampled            *logger.Logger
 	ingestionAccessLog accesslog.Log
 	pipeline           queue.Client
 	broadcaster        queue.Client
-	writeTimeout       time.Duration
+	*discoveryService
+	sampled      *logger.Logger
+	metrics      *metrics
+	writeTimeout time.Duration
 }
 
 func (s *streamService) setLogger(log *logger.Logger) {
@@ -63,12 +66,23 @@ func (s *streamService) activeIngestionAccessLog(root string) (err error) {
 
 func (s *streamService) Write(stream streamv1.StreamService_WriteServer) error {
 	reply := func(metadata *commonv1.Metadata, status modelv1.Status, messageId uint64, stream streamv1.StreamService_WriteServer, logger *logger.Logger) {
+		if status != modelv1.Status_STATUS_SUCCEED {
+			s.metrics.totalStreamMsgReceivedErr.Inc(1, metadata.Group, "stream", "write")
+		}
+		s.metrics.totalStreamMsgReceived.Inc(1, metadata.Group, "stream", "write")
 		if errResp := stream.Send(&streamv1.WriteResponse{Metadata: metadata, Status: status, MessageId: messageId}); errResp != nil {
-			logger.Debug().Err(errResp).Msg("failed to send response")
+			logger.Debug().Err(errResp).Msg("failed to send stream write response")
+			s.metrics.totalStreamMsgSentErr.Inc(1, metadata.Group, "stream", "write")
 		}
 	}
+	s.metrics.totalStreamStarted.Inc(1, "stream", "write")
 	publisher := s.pipeline.NewBatchPublisher(s.writeTimeout)
-	defer publisher.Close()
+	start := time.Now()
+	defer func() {
+		publisher.Close()
+		s.metrics.totalStreamFinished.Inc(1, "stream", "write")
+		s.metrics.totalStreamLatency.Inc(time.Since(start).Seconds(), "stream", "write")
+	}()
 	ctx := stream.Context()
 	for {
 		select {
@@ -81,9 +95,12 @@ func (s *streamService) Write(stream streamv1.StreamService_WriteServer) error {
 			return nil
 		}
 		if err != nil {
-			s.sampled.Error().Stringer("written", writeEntity).Err(err).Msg("failed to receive message")
+			if !errors.Is(err, context.DeadlineExceeded) && !errors.Is(err, context.Canceled) {
+				s.sampled.Error().Stringer("written", writeEntity).Err(err).Msg("failed to receive message")
+			}
 			return err
 		}
+		s.metrics.totalStreamMsgReceived.Inc(1, writeEntity.Metadata.Group, "stream", "write")
 		if errTime := timestamp.CheckPb(writeEntity.GetElement().Timestamp); errTime != nil {
 			s.sampled.Error().Stringer("written", writeEntity).Err(errTime).Msg("the element time is invalid")
 			reply(nil, modelv1.Status_STATUS_INVALID_TIMESTAMP, writeEntity.GetMessageId(), stream, s.sampled)
@@ -114,10 +131,9 @@ func (s *streamService) Write(stream streamv1.StreamService_WriteServer) error {
 			}
 		}
 		iwr := &streamv1.InternalWriteRequest{
-			Request:    writeEntity,
-			ShardId:    uint32(shardID),
-			SeriesHash: pbv1.HashEntity(entity),
-			// TODO: remove the first value (stream name) of tagValues
+			Request:      writeEntity,
+			ShardId:      uint32(shardID),
+			SeriesHash:   pbv1.HashEntity(entity),
 			EntityValues: tagValues[1:].Encode(),
 		}
 		nodeID, errPickNode := s.nodeRegistry.Locate(writeEntity.GetMetadata().GetGroup(), writeEntity.GetMetadata().GetName(), uint32(shardID))
@@ -127,28 +143,56 @@ func (s *streamService) Write(stream streamv1.StreamService_WriteServer) error {
 			continue
 		}
 		message := bus.NewBatchMessageWithNode(bus.MessageID(time.Now().UnixNano()), nodeID, iwr)
-		_, errWritePub := publisher.Publish(data.TopicStreamWrite, message)
+		_, errWritePub := publisher.Publish(ctx, data.TopicStreamWrite, message)
 		if errWritePub != nil {
 			s.sampled.Error().Err(errWritePub).RawJSON("written", logger.Proto(writeEntity)).Str("nodeID", nodeID).Msg("failed to send a message")
 			reply(writeEntity.GetMetadata(), modelv1.Status_STATUS_INTERNAL_ERROR, writeEntity.GetMessageId(), stream, s.sampled)
 			continue
 		}
-		reply(nil, modelv1.Status_STATUS_SUCCEED, writeEntity.GetMessageId(), stream, s.sampled)
+		reply(writeEntity.GetMetadata(), modelv1.Status_STATUS_SUCCEED, writeEntity.GetMessageId(), stream, s.sampled)
 	}
 }
 
 var emptyStreamQueryResponse = &streamv1.QueryResponse{Elements: make([]*streamv1.Element, 0)}
 
-func (s *streamService) Query(_ context.Context, req *streamv1.QueryRequest) (*streamv1.QueryResponse, error) {
+func (s *streamService) Query(ctx context.Context, req *streamv1.QueryRequest) (resp *streamv1.QueryResponse, err error) {
+	for _, g := range req.Groups {
+		s.metrics.totalStarted.Inc(1, g, "stream", "query")
+	}
+	start := time.Now()
+	defer func() {
+		for _, g := range req.Groups {
+			s.metrics.totalFinished.Inc(1, g, "stream", "query")
+			if err != nil {
+				s.metrics.totalErr.Inc(1, g, "stream", "query")
+			}
+			s.metrics.totalLatency.Inc(time.Since(start).Seconds(), g, "stream", "query")
+		}
+	}()
 	timeRange := req.GetTimeRange()
 	if timeRange == nil {
 		req.TimeRange = timestamp.DefaultTimeRange
 	}
-	if err := timestamp.CheckTimeRange(req.GetTimeRange()); err != nil {
+	if err = timestamp.CheckTimeRange(req.GetTimeRange()); err != nil {
 		return nil, status.Errorf(codes.InvalidArgument, "%v is invalid :%s", req.GetTimeRange(), err)
 	}
-	message := bus.NewMessage(bus.MessageID(time.Now().UnixNano()), req)
-	feat, errQuery := s.broadcaster.Publish(data.TopicStreamQuery, message)
+	now := time.Now()
+	if req.Trace {
+		tracer, _ := query.NewTracer(ctx, now.Format(time.RFC3339Nano))
+		span, _ := tracer.StartSpan(ctx, "stream-grpc")
+		span.Tag("request", convert.BytesToString(logger.Proto(req)))
+		defer func() {
+			if err != nil {
+				span.Error(err)
+			} else {
+				span.AddSubTrace(resp.Trace)
+				resp.Trace = tracer.ToProto()
+			}
+			span.Stop()
+		}()
+	}
+	message := bus.NewMessage(bus.MessageID(now.UnixNano()), req)
+	feat, errQuery := s.broadcaster.Publish(ctx, data.TopicStreamQuery, message)
 	if errQuery != nil {
 		if errors.Is(errQuery, io.EOF) {
 			return emptyStreamQueryResponse, nil
