@@ -18,8 +18,9 @@
 package measure
 
 import (
+	"fmt"
+	"slices"
 	"sort"
-	"sync"
 
 	"github.com/apache/skywalking-banyandb/api/common"
 	modelv1 "github.com/apache/skywalking-banyandb/api/proto/banyandb/model/v1"
@@ -28,12 +29,16 @@ import (
 	"github.com/apache/skywalking-banyandb/pkg/fs"
 	"github.com/apache/skywalking-banyandb/pkg/logger"
 	pbv1 "github.com/apache/skywalking-banyandb/pkg/pb/v1"
+	"github.com/apache/skywalking-banyandb/pkg/pool"
+	"github.com/apache/skywalking-banyandb/pkg/query/model"
 	"github.com/apache/skywalking-banyandb/pkg/timestamp"
 )
 
 type block struct {
 	timestamps []int64
 
+	versions []int64
+
 	tagFamilies []columnFamily
 
 	field columnFamily
@@ -41,6 +46,7 @@ type block struct {
 
 func (b *block) reset() {
 	b.timestamps = b.timestamps[:0]
+	b.versions = b.versions[:0]
 
 	tff := b.tagFamilies
 	for i := range tff {
@@ -50,7 +56,7 @@ func (b *block) reset() {
 	b.field.reset()
 }
 
-func (b *block) mustInitFromDataPoints(timestamps []int64, tagFamilies [][]nameValues, fields []nameValues) {
+func (b *block) mustInitFromDataPoints(timestamps []int64, versions []int64, tagFamilies [][]nameValues, fields []nameValues) {
 	b.reset()
 	size := len(timestamps)
 	if size == 0 {
@@ -65,13 +71,14 @@ func (b *block) mustInitFromDataPoints(timestamps []int64, tagFamilies [][]nameV
 
 	assertTimestampsSorted(timestamps)
 	b.timestamps = append(b.timestamps, timestamps...)
+	b.versions = append(b.versions, versions...)
 	b.mustInitFromTagsAndFields(tagFamilies, fields)
 }
 
 func assertTimestampsSorted(timestamps []int64) {
 	for i := range timestamps {
 		if i > 0 && timestamps[i-1] > timestamps[i] {
-			logger.Panicf("log entries must be sorted by timestamp; got the previous entry with bigger timestamp %d than the current entry with timestamp %d",
+			logger.Panicf("data points must be sorted by timestamp; got the previous data point with bigger timestamp %d than the current data point with timestamp %d",
 				timestamps[i-1], timestamps[i])
 		}
 	}
@@ -136,7 +143,7 @@ func (b *block) mustWriteTo(sid common.SeriesID, bm *blockMetadata, ww *writers)
 	bm.uncompressedSizeBytes = b.uncompressedSizeBytes()
 	bm.count = uint64(b.Len())
 
-	mustWriteTimestampsTo(&bm.timestamps, b.timestamps, &ww.timestampsWriter)
+	mustWriteTimestampsTo(&bm.timestamps, b.timestamps, b.versions, &ww.timestampsWriter)
 
 	for ti := range b.tagFamilies {
 		b.marshalTagFamily(b.tagFamilies[ti], bm, ww)
@@ -252,7 +259,7 @@ func (b *block) unmarshalTagFamilyFromSeqReaders(decoder *encoding.BytesBlockDec
 func (b *block) uncompressedSizeBytes() uint64 {
 	dataPointsCount := uint64(b.Len())
 
-	n := dataPointsCount * 8
+	n := dataPointsCount * (8 + 8) // 8 bytes for timestamp and 8 bytes for version
 
 	tff := b.tagFamilies
 	for i := range tff {
@@ -284,7 +291,7 @@ func (b *block) uncompressedSizeBytes() uint64 {
 func (b *block) mustReadFrom(decoder *encoding.BytesBlockDecoder, p *part, bm blockMetadata) {
 	b.reset()
 
-	b.timestamps = mustReadTimestampsFrom(b.timestamps, &bm.timestamps, int(bm.count), p.timestamps)
+	b.timestamps, b.versions = mustReadTimestampsFrom(b.timestamps, b.versions, &bm.timestamps, int(bm.count), p.timestamps)
 
 	cc := b.field.resizeColumns(len(bm.field.columnMetadata))
 	for i := range cc {
@@ -307,7 +314,7 @@ func (b *block) mustReadFrom(decoder *encoding.BytesBlockDecoder, p *part, bm bl
 func (b *block) mustSeqReadFrom(decoder *encoding.BytesBlockDecoder, seqReaders *seqReaders, bm blockMetadata) {
 	b.reset()
 
-	b.timestamps = mustSeqReadTimestampsFrom(b.timestamps, &bm.timestamps, int(bm.count), &seqReaders.timestamps)
+	b.timestamps, b.versions = mustSeqReadTimestampsFrom(b.timestamps, b.versions, &bm.timestamps, int(bm.count), &seqReaders.timestamps)
 
 	cc := b.field.resizeColumns(len(bm.field.columnMetadata))
 	for i := range cc {
@@ -333,35 +340,35 @@ func (b *block) sortTagFamilies() {
 	})
 }
 
-func mustWriteTimestampsTo(tm *timestampsMetadata, timestamps []int64, timestampsWriter *writer) {
+func mustWriteTimestampsTo(tm *timestampsMetadata, timestamps, versions []int64, timestampsWriter *writer) {
 	tm.reset()
 
 	bb := bigValuePool.Generate()
 	defer bigValuePool.Release(bb)
 	bb.Buf, tm.encodeType, tm.min = encoding.Int64ListToBytes(bb.Buf[:0], timestamps)
-	if len(bb.Buf) > maxTimestampsBlockSize {
-		logger.Panicf("too big block with timestamps: %d bytes; the maximum supported size is %d bytes", len(bb.Buf), maxTimestampsBlockSize)
+	tm.encodeType = encoding.GetVersionType(tm.encodeType)
+	if tm.encodeType == encoding.EncodeTypeUnknown {
+		logger.Panicf("unexpected encodeType %d", tm.encodeType)
+		return
 	}
 	tm.max = timestamps[len(timestamps)-1]
 	tm.offset = timestampsWriter.bytesWritten
-	tm.size = uint64(len(bb.Buf))
+	tm.versionOffset = uint64(len(bb.Buf))
+	timestampsWriter.MustWrite(bb.Buf)
+	bb.Buf, tm.versionEncodeType, tm.versionFirst = encoding.Int64ListToBytes(bb.Buf[:0], versions)
+	tm.size = tm.versionOffset + uint64(len(bb.Buf))
 	timestampsWriter.MustWrite(bb.Buf)
 }
 
-func mustReadTimestampsFrom(dst []int64, tm *timestampsMetadata, count int, reader fs.Reader) []int64 {
+func mustReadTimestampsFrom(timestamps, versions []int64, tm *timestampsMetadata, count int, reader fs.Reader) ([]int64, []int64) {
 	bb := bigValuePool.Generate()
 	defer bigValuePool.Release(bb)
 	bb.Buf = bytes.ResizeExact(bb.Buf, int(tm.size))
 	fs.MustReadData(reader, int64(tm.offset), bb.Buf)
-	var err error
-	dst, err = encoding.BytesToInt64List(dst, bb.Buf, tm.encodeType, tm.min, count)
-	if err != nil {
-		logger.Panicf("%s: cannot unmarshal timestamps: %v", reader.Path(), err)
-	}
-	return dst
+	return mustDecodeTimestampsWithVersions(timestamps, versions, tm, count, reader.Path(), bb.Buf)
 }
 
-func mustSeqReadTimestampsFrom(dst []int64, tm *timestampsMetadata, count int, reader *seqReader) []int64 {
+func mustSeqReadTimestampsFrom(timestamps, versions []int64, tm *timestampsMetadata, count int, reader *seqReader) ([]int64, []int64) {
 	if tm.offset != reader.bytesRead {
 		logger.Panicf("offset %d must be equal to bytesRead %d", tm.offset, reader.bytesRead)
 	}
@@ -369,12 +376,27 @@ func mustSeqReadTimestampsFrom(dst []int64, tm *timestampsMetadata, count int, r
 	defer bigValuePool.Release(bb)
 	bb.Buf = bytes.ResizeExact(bb.Buf, int(tm.size))
 	reader.mustReadFull(bb.Buf)
+	return mustDecodeTimestampsWithVersions(timestamps, versions, tm, count, reader.Path(), bb.Buf)
+}
+
+func mustDecodeTimestampsWithVersions(timestamps, versions []int64, tm *timestampsMetadata, count int, path string, src []byte) ([]int64, []int64) {
 	var err error
-	dst, err = encoding.BytesToInt64List(dst, bb.Buf, tm.encodeType, tm.min, count)
-	if err != nil {
-		logger.Panicf("%s: cannot unmarshal timestamps: %v", reader.Path(), err)
+	t := encoding.GetCommonType(tm.encodeType)
+	if t == encoding.EncodeTypeUnknown {
+		logger.Panicf("unexpected encodeType %d", tm.encodeType)
 	}
-	return dst
+	if tm.size < tm.versionOffset {
+		logger.Panicf("size %d must be greater than versionOffset %d", tm.size, tm.versionOffset)
+	}
+	timestamps, err = encoding.BytesToInt64List(timestamps, src[:tm.versionOffset], t, tm.min, count)
+	if err != nil {
+		logger.Panicf("%s: cannot unmarshal timestamps with versions: %v", path, err)
+	}
+	versions, err = encoding.BytesToInt64List(versions, src[tm.versionOffset:], tm.versionEncodeType, tm.versionFirst, count)
+	if err != nil {
+		logger.Panicf("%s: cannot unmarshal versions: %v", path, err)
+	}
+	return timestamps, versions
 }
 
 func generateBlock() *block {
@@ -382,7 +404,7 @@ func generateBlock() *block {
 	if v == nil {
 		return &block{}
 	}
-	return v.(*block)
+	return v
 }
 
 func releaseBlock(b *block) {
@@ -390,15 +412,16 @@ func releaseBlock(b *block) {
 	blockPool.Put(b)
 }
 
-var blockPool sync.Pool
+var blockPool = pool.Register[*block]("measure-block")
 
 type blockCursor struct {
 	p                   *part
 	fields              columnFamily
 	timestamps          []int64
+	versions            []int64
 	tagFamilies         []columnFamily
 	columnValuesDecoder encoding.BytesBlockDecoder
-	tagProjection       []pbv1.TagProjection
+	tagProjection       []model.TagProjection
 	fieldProjection     []string
 	bm                  blockMetadata
 	idx                 int
@@ -416,6 +439,7 @@ func (bc *blockCursor) reset() {
 	bc.fieldProjection = bc.fieldProjection[:0]
 
 	bc.timestamps = bc.timestamps[:0]
+	bc.versions = bc.versions[:0]
 
 	tff := bc.tagFamilies
 	for i := range tff {
@@ -435,8 +459,8 @@ func (bc *blockCursor) init(p *part, bm *blockMetadata, queryOpts queryOptions)
 	bc.fieldProjection = queryOpts.FieldProjection
 }
 
-func (bc *blockCursor) copyAllTo(r *pbv1.MeasureResult, entityValuesAll map[common.SeriesID]map[string]*modelv1.TagValue,
-	tagProjection []pbv1.TagProjection, desc bool,
+func (bc *blockCursor) copyAllTo(r *model.MeasureResult, storedIndexValue map[common.SeriesID]map[string]*modelv1.TagValue,
+	tagProjection []model.TagProjection, desc bool,
 ) {
 	var idx, offset int
 	if desc {
@@ -452,24 +476,29 @@ func (bc *blockCursor) copyAllTo(r *pbv1.MeasureResult, entityValuesAll map[comm
 	size := offset - idx
 	r.SID = bc.bm.seriesID
 	r.Timestamps = append(r.Timestamps, bc.timestamps[idx:offset]...)
-	var entityValues map[string]*modelv1.TagValue
-	if entityValuesAll != nil {
-		entityValues = entityValuesAll[r.SID]
+	r.Versions = append(r.Versions, bc.versions[idx:offset]...)
+	if desc {
+		slices.Reverse(r.Timestamps)
+		slices.Reverse(r.Versions)
+	}
+	var indexValue map[string]*modelv1.TagValue
+	if storedIndexValue != nil {
+		indexValue = storedIndexValue[r.SID]
 	}
 OUTER:
 	for _, tp := range tagProjection {
-		tf := pbv1.TagFamily{
+		tf := model.TagFamily{
 			Name: tp.Family,
 		}
 		var cf *columnFamily
 		for _, tagName := range tp.Names {
-			t := pbv1.Tag{
+			t := model.Tag{
 				Name: tagName,
 			}
-			if entityValues != nil && entityValues[tagName] != nil {
+			if indexValue != nil && indexValue[tagName] != nil {
 				t.Values = make([]*modelv1.TagValue, size)
 				for i := 0; i < size; i++ {
-					t.Values[i] = entityValues[tagName]
+					t.Values[i] = indexValue[tagName]
 				}
 				tf.Tags = append(tf.Tags, t)
 				continue
@@ -484,7 +513,7 @@ OUTER:
 			}
 			if cf == nil {
 				for _, n := range tp.Names {
-					t = pbv1.Tag{
+					t = model.Tag{
 						Name:   n,
 						Values: make([]*modelv1.TagValue, size),
 					}
@@ -511,38 +540,44 @@ OUTER:
 				for i := 0; i < size; i++ {
 					t.Values[i] = pbv1.NullTagValue
 				}
+			} else if desc {
+				slices.Reverse(t.Values)
 			}
 			tf.Tags = append(tf.Tags, t)
 		}
 		r.TagFamilies = append(r.TagFamilies, tf)
 	}
 	for _, c := range bc.fields.columns {
-		f := pbv1.Field{
+		f := model.Field{
 			Name: c.name,
 		}
 		for _, v := range c.values[idx:offset] {
 			f.Values = append(f.Values, mustDecodeFieldValue(c.valueType, v))
 		}
+		if desc {
+			slices.Reverse(f.Values)
+		}
 		r.Fields = append(r.Fields, f)
 	}
 }
 
-func (bc *blockCursor) copyTo(r *pbv1.MeasureResult, entityValuesAll map[common.SeriesID]map[string]*modelv1.TagValue,
-	tagProjection []pbv1.TagProjection,
+func (bc *blockCursor) copyTo(r *model.MeasureResult, storedIndexValue map[common.SeriesID]map[string]*modelv1.TagValue,
+	tagProjection []model.TagProjection,
 ) {
 	r.SID = bc.bm.seriesID
 	r.Timestamps = append(r.Timestamps, bc.timestamps[bc.idx])
-	var entityValues map[string]*modelv1.TagValue
-	if entityValuesAll != nil {
-		entityValues = entityValuesAll[r.SID]
+	r.Versions = append(r.Versions, bc.versions[bc.idx])
+	var indexValue map[string]*modelv1.TagValue
+	if storedIndexValue != nil {
+		indexValue = storedIndexValue[r.SID]
 	}
 	if len(r.TagFamilies) == 0 {
 		for _, tp := range tagProjection {
-			tf := pbv1.TagFamily{
+			tf := model.TagFamily{
 				Name: tp.Family,
 			}
 			for _, n := range tp.Names {
-				t := pbv1.Tag{
+				t := model.Tag{
 					Name: n,
 				}
 				tf.Tags = append(tf.Tags, t)
@@ -555,8 +590,8 @@ func (bc *blockCursor) copyTo(r *pbv1.MeasureResult, entityValuesAll map[common.
 		var cf *columnFamily
 		for j := range r.TagFamilies[i].Tags {
 			tagName := r.TagFamilies[i].Tags[j].Name
-			if entityValues != nil && entityValues[tagName] != nil {
-				r.TagFamilies[i].Tags[j].Values = append(r.TagFamilies[i].Tags[j].Values, entityValues[tagName])
+			if indexValue != nil && indexValue[tagName] != nil {
+				r.TagFamilies[i].Tags[j].Values = append(r.TagFamilies[i].Tags[j].Values, indexValue[tagName])
 				continue
 			}
 			if cf == nil {
@@ -587,7 +622,7 @@ func (bc *blockCursor) copyTo(r *pbv1.MeasureResult, entityValuesAll map[common.
 
 	if len(r.Fields) == 0 {
 		for _, n := range bc.fieldProjection {
-			f := pbv1.Field{
+			f := model.Field{
 				Name: n,
 			}
 			r.Fields = append(r.Fields, f)
@@ -598,6 +633,44 @@ func (bc *blockCursor) copyTo(r *pbv1.MeasureResult, entityValuesAll map[common.
 	}
 }
 
+func (bc *blockCursor) replace(r *model.MeasureResult, storedIndexValue map[common.SeriesID]map[string]*modelv1.TagValue) {
+	r.SID = bc.bm.seriesID
+	r.Timestamps[len(r.Timestamps)-1] = bc.timestamps[bc.idx]
+	r.Versions[len(r.Versions)-1] = bc.versions[bc.idx]
+	var indexValue map[string]*modelv1.TagValue
+	if storedIndexValue != nil {
+		indexValue = storedIndexValue[r.SID]
+	}
+	for i := range r.TagFamilies {
+		tfName := r.TagFamilies[i].Name
+		var cf *columnFamily
+		for j := range r.TagFamilies[i].Tags {
+			tagName := r.TagFamilies[i].Tags[j].Name
+			if indexValue != nil && indexValue[tagName] != nil {
+				r.TagFamilies[i].Tags[j].Values[len(r.TagFamilies[i].Tags[j].Values)-1] = indexValue[tagName]
+				continue
+			}
+			if cf == nil {
+				for i := range bc.tagFamilies {
+					if bc.tagFamilies[i].name == tfName {
+						cf = &bc.tagFamilies[i]
+						break
+					}
+				}
+			}
+			for _, c := range cf.columns {
+				if c.name == tagName {
+					r.TagFamilies[i].Tags[j].Values[len(r.TagFamilies[i].Tags[j].Values)-1] = mustDecodeTagValue(c.valueType, c.values[bc.idx])
+					break
+				}
+			}
+		}
+	}
+	for i, c := range bc.fields.columns {
+		r.Fields[i].Values[len(r.Fields[i].Values)-1] = mustDecodeFieldValue(c.valueType, c.values[bc.idx])
+	}
+}
+
 func (bc *blockCursor) loadData(tmpBlock *block) bool {
 	tmpBlock.reset()
 	cfm := make([]columnMetadata, 0, len(bc.fieldProjection))
@@ -629,6 +702,7 @@ func (bc *blockCursor) loadData(tmpBlock *block) bool {
 		return false
 	}
 	bc.timestamps = append(bc.timestamps, tmpBlock.timestamps[start:end+1]...)
+	bc.versions = append(bc.versions, tmpBlock.versions[start:end+1]...)
 
 	for _, cf := range tmpBlock.tagFamilies {
 		tf := columnFamily{
@@ -670,14 +744,14 @@ func (bc *blockCursor) loadData(tmpBlock *block) bool {
 	return true
 }
 
-var blockCursorPool sync.Pool
+var blockCursorPool = pool.Register[*blockCursor]("measure-blockCursor")
 
 func generateBlockCursor() *blockCursor {
 	v := blockCursorPool.Get()
 	if v == nil {
 		return &blockCursor{}
 	}
-	return v.(*blockCursor)
+	return v
 }
 
 func releaseBlockCursor(bc *blockCursor) {
@@ -687,9 +761,8 @@ func releaseBlockCursor(bc *blockCursor) {
 
 type blockPointer struct {
 	block
-	bm         blockMetadata
-	idx        int
-	lastPartID uint64
+	bm  blockMetadata
+	idx int
 }
 
 func (bi *blockPointer) updateMetadata() {
@@ -703,7 +776,7 @@ func (bi *blockPointer) updateMetadata() {
 }
 
 func (bi *blockPointer) copyFrom(src *blockPointer) {
-	bi.idx = 0
+	bi.reset()
 	bi.bm.copyFrom(&src.bm)
 	bi.appendAll(src)
 }
@@ -715,63 +788,206 @@ func (bi *blockPointer) appendAll(b *blockPointer) {
 	bi.append(b, len(b.timestamps))
 }
 
+var log = logger.GetLogger("measure").Named("block")
+
 func (bi *blockPointer) append(b *blockPointer, offset int) {
 	if offset <= b.idx {
 		return
 	}
 	if len(bi.tagFamilies) == 0 && len(b.tagFamilies) > 0 {
-		for _, tf := range b.tagFamilies {
-			tagFamily := columnFamily{name: tf.name}
-			for _, c := range tf.columns {
-				col := column{name: c.name, valueType: c.valueType}
-				assertIdxAndOffset(col.name, len(c.values), b.idx, offset)
-				col.values = append(col.values, c.values[b.idx:offset]...)
-				tagFamily.columns = append(tagFamily.columns, col)
-			}
-			bi.tagFamilies = append(bi.tagFamilies, tagFamily)
-		}
+		fullTagAppend(bi, b, offset)
 	} else {
-		if len(bi.tagFamilies) != len(b.tagFamilies) {
-			logger.Panicf("unexpected number of tag families: got %d; want %d", len(bi.tagFamilies), len(b.tagFamilies))
-		}
-		for i := range bi.tagFamilies {
-			if bi.tagFamilies[i].name != b.tagFamilies[i].name {
-				logger.Panicf("unexpected tag family name: got %q; want %q", bi.tagFamilies[i].name, b.tagFamilies[i].name)
-			}
-			if len(bi.tagFamilies[i].columns) != len(b.tagFamilies[i].columns) {
-				logger.Panicf("unexpected number of tags for tag family %q: got %d; want %d", bi.tagFamilies[i].name, len(bi.tagFamilies[i].columns), len(b.tagFamilies[i].columns))
-			}
-			for j := range bi.tagFamilies[i].columns {
-				if bi.tagFamilies[i].columns[j].name != b.tagFamilies[i].columns[j].name {
-					logger.Panicf("unexpected tag name for tag family %q: got %q; want %q", bi.tagFamilies[i].name, bi.tagFamilies[i].columns[j].name, b.tagFamilies[i].columns[j].name)
-				}
-				assertIdxAndOffset(b.tagFamilies[i].columns[j].name, len(b.tagFamilies[i].columns[j].values), b.idx, offset)
-				bi.tagFamilies[i].columns[j].values = append(bi.tagFamilies[i].columns[j].values, b.tagFamilies[i].columns[j].values[b.idx:offset]...)
+		if err := fastTagAppend(bi, b, offset); err != nil {
+			if log.Debug().Enabled() {
+				log.Debug().Msgf("fastTagMerge failed: %v; falling back to fullTagMerge", err)
 			}
+			fullTagAppend(bi, b, offset)
 		}
 	}
 
 	if len(bi.field.columns) == 0 && len(b.field.columns) > 0 {
-		for _, c := range b.field.columns {
-			col := column{name: c.name, valueType: c.valueType}
-			assertIdxAndOffset(col.name, len(c.values), b.idx, offset)
-			col.values = append(col.values, c.values[b.idx:offset]...)
-			bi.field.columns = append(bi.field.columns, col)
-		}
+		fullFieldAppend(bi, b, offset)
 	} else {
-		if len(bi.field.columns) != len(b.field.columns) {
-			logger.Panicf("unexpected number of fields: got %d; want %d", len(bi.field.columns), len(b.field.columns))
-		}
-		for i := range bi.field.columns {
-			assertIdxAndOffset(b.field.columns[i].name, len(b.field.columns[i].values), b.idx, offset)
-			bi.field.columns[i].values = append(bi.field.columns[i].values, b.field.columns[i].values[b.idx:offset]...)
+		if err := fastFieldAppend(bi, b, offset); err != nil {
+			if log.Debug().Enabled() {
+				log.Debug().Msgf("fastFieldAppend failed: %v; falling back to fullFieldAppend", err)
+			}
+			fullFieldAppend(bi, b, offset)
 		}
 	}
 
 	assertIdxAndOffset("timestamps", len(b.timestamps), bi.idx, offset)
 	bi.timestamps = append(bi.timestamps, b.timestamps[b.idx:offset]...)
+	assertIdxAndOffset("versions", len(b.versions), bi.idx, offset)
+	bi.versions = append(bi.versions, b.versions[b.idx:offset]...)
+}
 
-	bi.lastPartID = b.lastPartID
+func fastTagAppend(bi, b *blockPointer, offset int) error {
+	if len(bi.tagFamilies) != len(b.tagFamilies) {
+		return fmt.Errorf("unexpected number of tag families: got %d; want %d", len(b.tagFamilies), len(bi.tagFamilies))
+	}
+	for i := range bi.tagFamilies {
+		if bi.tagFamilies[i].name != b.tagFamilies[i].name {
+			return fmt.Errorf("unexpected tag family name: got %q; want %q", b.tagFamilies[i].name, bi.tagFamilies[i].name)
+		}
+		if len(bi.tagFamilies[i].columns) != len(b.tagFamilies[i].columns) {
+			return fmt.Errorf("unexpected number of tags for tag family %q: got %d; want %d",
+				bi.tagFamilies[i].name, len(b.tagFamilies[i].columns), len(bi.tagFamilies[i].columns))
+		}
+		for j := range bi.tagFamilies[i].columns {
+			if bi.tagFamilies[i].columns[j].name != b.tagFamilies[i].columns[j].name {
+				return fmt.Errorf("unexpected tag name for tag family %q: got %q; want %q",
+					bi.tagFamilies[i].name, b.tagFamilies[i].columns[j].name, bi.tagFamilies[i].columns[j].name)
+			}
+			assertIdxAndOffset(b.tagFamilies[i].columns[j].name, len(b.tagFamilies[i].columns[j].values), b.idx, offset)
+			bi.tagFamilies[i].columns[j].values = append(bi.tagFamilies[i].columns[j].values, b.tagFamilies[i].columns[j].values[b.idx:offset]...)
+		}
+	}
+	return nil
+}
+
+func fullTagAppend(bi, b *blockPointer, offset int) {
+	existDataSize := len(bi.timestamps)
+	appendTagFamilies := func(tf columnFamily) {
+		tagFamily := columnFamily{name: tf.name}
+		for i := range tf.columns {
+			assertIdxAndOffset(tf.columns[i].name, len(tf.columns[i].values), b.idx, offset)
+			col := column{name: tf.columns[i].name, valueType: tf.columns[i].valueType}
+			for j := 0; j < existDataSize; j++ {
+				col.values = append(col.values, nil)
+			}
+			col.values = append(col.values, tf.columns[i].values[b.idx:offset]...)
+			tagFamily.columns = append(tagFamily.columns, col)
+		}
+		bi.tagFamilies = append(bi.tagFamilies, tagFamily)
+	}
+	if len(bi.tagFamilies) == 0 {
+		for _, tf := range b.tagFamilies {
+			appendTagFamilies(tf)
+		}
+		return
+	}
+
+	tagFamilyMap := make(map[string]*columnFamily)
+	for i := range bi.tagFamilies {
+		tagFamilyMap[bi.tagFamilies[i].name] = &bi.tagFamilies[i]
+	}
+
+	for _, tf := range b.tagFamilies {
+		if existingTagFamily, exists := tagFamilyMap[tf.name]; exists {
+			columnMap := make(map[string]*column)
+			for i := range existingTagFamily.columns {
+				columnMap[existingTagFamily.columns[i].name] = &existingTagFamily.columns[i]
+			}
+
+			for _, c := range tf.columns {
+				if existingColumn, exists := columnMap[c.name]; exists {
+					assertIdxAndOffset(c.name, len(c.values), b.idx, offset)
+					existingColumn.values = append(existingColumn.values, c.values[b.idx:offset]...)
+				} else {
+					assertIdxAndOffset(c.name, len(c.values), b.idx, offset)
+					col := column{name: c.name, valueType: c.valueType}
+					for j := 0; j < existDataSize; j++ {
+						col.values = append(col.values, nil)
+					}
+					col.values = append(col.values, c.values[b.idx:offset]...)
+					existingTagFamily.columns = append(existingTagFamily.columns, col)
+				}
+			}
+		} else {
+			appendTagFamilies(tf)
+		}
+	}
+	for k := range tagFamilyMap {
+		delete(tagFamilyMap, k)
+	}
+	for i := range b.tagFamilies {
+		tagFamilyMap[b.tagFamilies[i].name] = &b.tagFamilies[i]
+	}
+	emptySize := offset - b.idx
+	for _, tf := range bi.tagFamilies {
+		if _, exists := tagFamilyMap[tf.name]; !exists {
+			for i := range tf.columns {
+				for j := 0; j < emptySize; j++ {
+					tf.columns[i].values = append(tf.columns[i].values, nil)
+				}
+			}
+		} else {
+			existingTagFamily := tagFamilyMap[tf.name]
+			columnMap := make(map[string]*column)
+			for i := range existingTagFamily.columns {
+				columnMap[existingTagFamily.columns[i].name] = &existingTagFamily.columns[i]
+			}
+			for i := range tf.columns {
+				if _, exists := columnMap[tf.columns[i].name]; !exists {
+					for j := 0; j < emptySize; j++ {
+						tf.columns[i].values = append(tf.columns[i].values, nil)
+					}
+				}
+			}
+		}
+	}
+}
+
+func fastFieldAppend(bi, b *blockPointer, offset int) error {
+	if len(bi.field.columns) != len(b.field.columns) {
+		return fmt.Errorf("unexpected number of fields: got %d; want %d", len(bi.field.columns), len(b.field.columns))
+	}
+	for i := range bi.field.columns {
+		if bi.field.columns[i].name != b.field.columns[i].name {
+			return fmt.Errorf("unexpected field name: got %q; want %q", b.field.columns[i].name, bi.field.columns[i].name)
+		}
+		assertIdxAndOffset(b.field.columns[i].name, len(b.field.columns[i].values), b.idx, offset)
+		bi.field.columns[i].values = append(bi.field.columns[i].values, b.field.columns[i].values[b.idx:offset]...)
+	}
+	return nil
+}
+
+func fullFieldAppend(bi, b *blockPointer, offset int) {
+	existDataSize := len(bi.timestamps)
+	appendFields := func(c column) {
+		col := column{name: c.name, valueType: c.valueType}
+		for j := 0; j < existDataSize; j++ {
+			col.values = append(col.values, nil)
+		}
+		col.values = append(col.values, c.values[b.idx:offset]...)
+		bi.field.columns = append(bi.field.columns, col)
+	}
+	if len(bi.field.columns) == 0 {
+		for _, c := range b.field.columns {
+			appendFields(c)
+		}
+		return
+	}
+
+	fieldMap := make(map[string]*column)
+	for i := range bi.field.columns {
+		fieldMap[bi.field.columns[i].name] = &bi.field.columns[i]
+	}
+
+	for _, c := range b.field.columns {
+		if existingField, exists := fieldMap[c.name]; exists {
+			assertIdxAndOffset(c.name, len(c.values), b.idx, offset)
+			existingField.values = append(existingField.values, c.values[b.idx:offset]...)
+		} else {
+			appendFields(c)
+		}
+	}
+	for k := range fieldMap {
+		delete(fieldMap, k)
+	}
+	for i := range b.field.columns {
+		fieldMap[b.field.columns[i].name] = &b.field.columns[i]
+	}
+
+	emptySize := offset - b.idx
+	for i := range bi.field.columns {
+		if _, exists := fieldMap[bi.field.columns[i].name]; !exists {
+			for j := 0; j < emptySize; j++ {
+				bi.field.columns[i].values = append(bi.field.columns[i].values, nil)
+			}
+		}
+	}
 }
 
 func assertIdxAndOffset(name string, length int, idx int, offset int) {
@@ -789,9 +1005,8 @@ func (bi *blockPointer) isFull() bool {
 
 func (bi *blockPointer) reset() {
 	bi.idx = 0
-	bi.lastPartID = 0
 	bi.block.reset()
-	bi.bm = blockMetadata{}
+	bi.bm.reset()
 }
 
 func generateBlockPointer() *blockPointer {
@@ -799,7 +1014,7 @@ func generateBlockPointer() *blockPointer {
 	if v == nil {
 		return &blockPointer{}
 	}
-	return v.(*blockPointer)
+	return v
 }
 
 func releaseBlockPointer(bi *blockPointer) {
@@ -807,4 +1022,4 @@ func releaseBlockPointer(bi *blockPointer) {
 	blockPointerPool.Put(bi)
 }
 
-var blockPointerPool sync.Pool
+var blockPointerPool = pool.Register[*blockPointer]("measure-blockPointer")
