@@ -61,6 +61,7 @@ func Test_mergeTwoBlocks(t *testing.T) {
 			left: &blockPointer{
 				block: block{
 					timestamps: []int64{1, 2},
+					versions:   []int64{1, 4},
 					tagFamilies: []columnFamily{
 						{
 							name: "arrTag",
@@ -82,6 +83,7 @@ func Test_mergeTwoBlocks(t *testing.T) {
 			right: &blockPointer{
 				block: block{
 					timestamps: []int64{3, 4},
+					versions:   []int64{5, 6},
 					tagFamilies: []columnFamily{
 						{
 							name: "arrTag",
@@ -107,6 +109,7 @@ func Test_mergeTwoBlocks(t *testing.T) {
 			left: &blockPointer{
 				block: block{
 					timestamps: []int64{1, 3},
+					versions:   []int64{1, 5},
 					tagFamilies: []columnFamily{
 						{
 							name: "arrTag",
@@ -128,6 +131,7 @@ func Test_mergeTwoBlocks(t *testing.T) {
 			right: &blockPointer{
 				block: block{
 					timestamps: []int64{2, 4},
+					versions:   []int64{4, 6},
 					tagFamilies: []columnFamily{
 						{
 							name: "arrTag",
@@ -151,9 +155,9 @@ func Test_mergeTwoBlocks(t *testing.T) {
 		{
 			name: "Merge two non-empty blocks with duplicated timestamps",
 			left: &blockPointer{
-				lastPartID: 1, // the less partID will be skipped
 				block: block{
 					timestamps: []int64{1, 2, 3},
+					versions:   []int64{1, 2, 3},
 					tagFamilies: []columnFamily{
 						{
 							name: "arrTag",
@@ -177,9 +181,9 @@ func Test_mergeTwoBlocks(t *testing.T) {
 				},
 			},
 			right: &blockPointer{
-				lastPartID: 2, // the greater partID will be appended
 				block: block{
 					timestamps: []int64{2, 3, 4},
+					versions:   []int64{4, 5, 6},
 					tagFamilies: []columnFamily{
 						{
 							name: "arrTag",
@@ -201,7 +205,7 @@ func Test_mergeTwoBlocks(t *testing.T) {
 					},
 				},
 			},
-			want: &blockPointer{block: mergedBlock, lastPartID: 2, bm: blockMetadata{timestamps: timestampsMetadata{min: 1, max: 4}}},
+			want: &blockPointer{block: mergedBlock, bm: blockMetadata{timestamps: timestampsMetadata{min: 1, max: 4}}},
 		},
 	}
 
@@ -218,6 +222,7 @@ func Test_mergeTwoBlocks(t *testing.T) {
 
 var mergedBlock = block{
 	timestamps: []int64{1, 2, 3, 4},
+	versions:   []int64{1, 4, 5, 6},
 	tagFamilies: []columnFamily{
 		{
 			name: "arrTag",
@@ -255,43 +260,52 @@ func Test_mergeParts(t *testing.T) {
 			name:    "Test with single part",
 			dpsList: []*dataPoints{dpsTS1},
 			want: []blockMetadata{
-				{seriesID: 1, count: 1, uncompressedSizeBytes: 1676},
-				{seriesID: 2, count: 1, uncompressedSizeBytes: 55},
-				{seriesID: 3, count: 1, uncompressedSizeBytes: 24},
+				{seriesID: 1, count: 1, uncompressedSizeBytes: 1684},
+				{seriesID: 2, count: 1, uncompressedSizeBytes: 63},
+				{seriesID: 3, count: 1, uncompressedSizeBytes: 32},
 			},
 		},
 		{
 			name:    "Test with multiple parts with different ts",
 			dpsList: []*dataPoints{dpsTS1, dpsTS2, dpsTS2},
 			want: []blockMetadata{
-				{seriesID: 1, count: 2, uncompressedSizeBytes: 3352},
-				{seriesID: 2, count: 2, uncompressedSizeBytes: 110},
-				{seriesID: 3, count: 2, uncompressedSizeBytes: 48},
+				{seriesID: 1, count: 2, uncompressedSizeBytes: 3368},
+				{seriesID: 2, count: 2, uncompressedSizeBytes: 126},
+				{seriesID: 3, count: 2, uncompressedSizeBytes: 64},
 			},
 		},
 		{
 			name:    "Test with multiple parts with same ts",
-			dpsList: []*dataPoints{dpsTS1, dpsTS1, dpsTS1},
+			dpsList: []*dataPoints{dpsTS11, dpsTS1},
 			want: []blockMetadata{
-				{seriesID: 1, count: 1, uncompressedSizeBytes: 1676},
-				{seriesID: 2, count: 1, uncompressedSizeBytes: 55},
-				{seriesID: 3, count: 1, uncompressedSizeBytes: 24},
+				{seriesID: 1, count: 1, uncompressedSizeBytes: 1684},
+				{seriesID: 2, count: 1, uncompressedSizeBytes: 63},
+				{seriesID: 3, count: 1, uncompressedSizeBytes: 32},
 			},
 		},
 		{
 			name:    "Test with multiple parts with a large quantity of different ts",
 			dpsList: []*dataPoints{generateHugeDps(1, 5000, 1), generateHugeDps(5001, 10000, 2)},
 			want: []blockMetadata{
-				{seriesID: 1, count: 2500, uncompressedSizeBytes: 4190000},
-				{seriesID: 1, count: 2500, uncompressedSizeBytes: 4190000},
-				{seriesID: 1, count: 2500, uncompressedSizeBytes: 4190000},
-				{seriesID: 1, count: 2500, uncompressedSizeBytes: 4190000},
-				{seriesID: 1, count: 2500, uncompressedSizeBytes: 4190000},
-				{seriesID: 1, count: 2500, uncompressedSizeBytes: 4190000},
-				{seriesID: 1, count: 2500, uncompressedSizeBytes: 4190000},
-				{seriesID: 1, count: 1250, uncompressedSizeBytes: 2095000},
-				{seriesID: 2, count: 2, uncompressedSizeBytes: 110},
-				{seriesID: 3, count: 2, uncompressedSizeBytes: 48},
+				{seriesID: 1, count: 1265, uncompressedSizeBytes: 2130260},
+				{seriesID: 1, count: 1265, uncompressedSizeBytes: 2130260},
+				{seriesID: 1, count: 1265, uncompressedSizeBytes: 2130260},
+				{seriesID: 1, count: 2470, uncompressedSizeBytes: 4159480},
+				{seriesID: 1, count: 1265, uncompressedSizeBytes: 2130260},
+				{seriesID: 1, count: 1265, uncompressedSizeBytes: 2130260},
+				{seriesID: 1, count: 1205, uncompressedSizeBytes: 2029220},
+				{seriesID: 2, count: 2, uncompressedSizeBytes: 126},
+				{seriesID: 3, count: 2, uncompressedSizeBytes: 64},
+			},
+		},
+		{
+			name:    "Test with multiple parts with a large small quantity of different ts",
+			dpsList: []*dataPoints{generateHugeSmallDps(1, 5000, 1), generateHugeSmallDps(5001, 10000, 2)},
+			want: []blockMetadata{
+				{seriesID: 1, count: 8192, uncompressedSizeBytes: 262144},
+				{seriesID: 1, count: 1808, uncompressedSizeBytes: 57856},
+				{seriesID: 2, count: 2, uncompressedSizeBytes: 32},
+				{seriesID: 3, count: 2, uncompressedSizeBytes: 64},
 			},
 		},
 	}
