@@ -19,7 +19,6 @@ package measure
 
 import (
 	"context"
-	"math"
 	"path"
 
 	"github.com/pkg/errors"
@@ -53,14 +52,16 @@ type Service interface {
 var _ Service = (*service)(nil)
 
 type service struct {
-	schemaRepo    *schemaRepo
-	writeListener bus.MessageListener
-	metadata      metadata.Repo
-	pipeline      queue.Server
-	localPipeline queue.Queue
-	option        option
-	l             *logger.Logger
-	root          string
+	writeListener  bus.MessageListener
+	metadata       metadata.Repo
+	pipeline       queue.Server
+	localPipeline  queue.Queue
+	metricPipeline queue.Server
+	omr            observability.MetricsRegistry
+	schemaRepo     *schemaRepo
+	l              *logger.Logger
+	option         option
+	root           string
 }
 
 func (s *service) Measure(metadata *commonv1.Metadata) (Measure, error) {
@@ -80,7 +81,7 @@ func (s *service) FlagSet() *run.FlagSet {
 	flagS.StringVar(&s.root, "measure-root-path", "/tmp", "the root path of database")
 	flagS.DurationVar(&s.option.flushTimeout, "measure-flush-timeout", defaultFlushTimeout, "the memory data timeout of measure")
 	s.option.mergePolicy = newDefaultMergePolicy()
-	flagS.Uint64Var(&s.option.mergePolicy.maxFanOutSize, "max-fan-out-size", math.MaxUint64, "the upper bound of a single file size after merge")
+	flagS.VarP(&s.option.mergePolicy.maxFanOutSize, "measure-max-fan-out-size", "", "the upper bound of a single file size after merge of measure")
 	return flagS
 }
 
@@ -99,15 +100,26 @@ func (s *service) Role() databasev1.Role {
 	return databasev1.Role_ROLE_DATA
 }
 
-func (s *service) PreRun(_ context.Context) error {
+func (s *service) PreRun(ctx context.Context) error {
 	s.l = logger.GetLogger(s.Name())
 	path := path.Join(s.root, s.Name())
 	observability.UpdatePath(path)
 	s.localPipeline = queue.Local()
 	s.schemaRepo = newSchemaRepo(path, s)
-	// run a serial watcher
+
+	if err := s.createNativeObservabilityGroup(ctx); err != nil {
+		return err
+	}
 
 	s.writeListener = setUpWriteCallback(s.l, s.schemaRepo)
+	// only subscribe metricPipeline for data node
+	if s.metricPipeline != nil {
+		err := s.metricPipeline.Subscribe(data.TopicMeasureWrite, s.writeListener)
+		if err != nil {
+			s.l.Err(err).Msgf("Fail to subscribe metricPipeline, %v", err)
+			return err
+		}
+	}
 	err := s.pipeline.Subscribe(data.TopicMeasureWrite, s.writeListener)
 	if err != nil {
 		return err
@@ -125,9 +137,11 @@ func (s *service) GracefulStop() {
 }
 
 // NewService returns a new service.
-func NewService(_ context.Context, metadata metadata.Repo, pipeline queue.Server) (Service, error) {
+func NewService(_ context.Context, metadata metadata.Repo, pipeline queue.Server, metricPipeline queue.Server, omr observability.MetricsRegistry) (Service, error) {
 	return &service{
-		metadata: metadata,
-		pipeline: pipeline,
+		metadata:       metadata,
+		pipeline:       pipeline,
+		metricPipeline: metricPipeline,
+		omr:            omr,
 	}, nil
 }
