@@ -19,7 +19,6 @@ package v1
 
 import (
 	"sort"
-	"sync"
 
 	"github.com/pkg/errors"
 
@@ -38,6 +37,13 @@ type Series struct {
 	ID           common.SeriesID
 }
 
+// CopyTo copies the content of the series to the destination series.
+func (s *Series) CopyTo(dst *Series) {
+	dst.Subject = s.Subject
+	dst.EntityValues = make([]*modelv1.TagValue, len(s.EntityValues))
+	copy(dst.EntityValues, s.EntityValues)
+}
+
 // Marshal encodes series to internal Buffer and generates ID.
 func (s *Series) Marshal() error {
 	s.Buffer = marshalEntityValue(s.Buffer, convert.StringToBytes(s.Subject))
@@ -90,26 +96,6 @@ func (s *Series) reset() {
 	s.Buffer = s.Buffer[:0]
 }
 
-// SeriesPool is a pool of Series.
-type SeriesPool struct {
-	pool sync.Pool
-}
-
-// Generate creates a new Series or gets one from the pool.
-func (sp *SeriesPool) Generate() *Series {
-	sv := sp.pool.Get()
-	if sv == nil {
-		return &Series{}
-	}
-	return sv.(*Series)
-}
-
-// Release puts a Series back to the pool.
-func (sp *SeriesPool) Release(s *Series) {
-	s.reset()
-	sp.pool.Put(s)
-}
-
 // SeriesList is a collection of Series.
 type SeriesList []*Series
 
