@@ -148,9 +148,7 @@ func ParseTagValue(tagValue *modelv1.TagValue) (TagValue, error) {
 		}
 		return *fv, nil
 	case *modelv1.TagValue_BinaryData:
-		return newValue(x.BinaryData), nil
-	case *modelv1.TagValue_Id:
-		return newValue([]byte(x.Id.GetValue())), nil
+		return newValue(bytes.Clone(x.BinaryData)), nil
 	}
 	return TagValue{}, errUnsupportedTagForIndexField
 }
