@@ -18,6 +18,8 @@
 package measure
 
 import (
+	"bytes"
+	"container/list"
 	"context"
 	"fmt"
 	"time"
@@ -31,7 +33,9 @@ import (
 	"github.com/apache/skywalking-banyandb/pkg/bus"
 	"github.com/apache/skywalking-banyandb/pkg/convert"
 	"github.com/apache/skywalking-banyandb/pkg/iter/sort"
+	"github.com/apache/skywalking-banyandb/pkg/logger"
 	pbv1 "github.com/apache/skywalking-banyandb/pkg/pb/v1"
+	"github.com/apache/skywalking-banyandb/pkg/query"
 	"github.com/apache/skywalking-banyandb/pkg/query/executor"
 	"github.com/apache/skywalking-banyandb/pkg/query/logical"
 )
@@ -83,7 +87,7 @@ func (ud *unresolvedDistributed) Analyze(s logical.Schema) (logical.Plan, error)
 		Name:            ud.originalQuery.Name,
 		Groups:          ud.originalQuery.Groups,
 		Criteria:        ud.originalQuery.Criteria,
-		Limit:           limit,
+		Limit:           limit + ud.originalQuery.Offset,
 		OrderBy:         ud.originalQuery.OrderBy,
 	}
 	if ud.groupByEntity {
@@ -153,35 +157,54 @@ type distributedPlan struct {
 	maxDataPointsSize uint32
 }
 
-func (t *distributedPlan) Execute(ctx context.Context) (executor.MIterator, error) {
+func (t *distributedPlan) Execute(ctx context.Context) (mi executor.MIterator, err error) {
 	dctx := executor.FromDistributedExecutionContext(ctx)
-	query := proto.Clone(t.queryTemplate).(*measurev1.QueryRequest)
-	query.TimeRange = dctx.TimeRange()
+	queryRequest := proto.Clone(t.queryTemplate).(*measurev1.QueryRequest)
+	queryRequest.TimeRange = dctx.TimeRange()
 	if t.maxDataPointsSize > 0 {
-		query.Limit = t.maxDataPointsSize
+		queryRequest.Limit = t.maxDataPointsSize
 	}
-	var allErr error
-	ff, err := dctx.Broadcast(defaultQueryTimeout, data.TopicMeasureQuery, bus.NewMessage(bus.MessageID(dctx.TimeRange().Begin.Nanos), query))
+	tracer := query.GetTracer(ctx)
+	var span *query.Span
+	if tracer != nil {
+		span, _ = tracer.StartSpan(ctx, "distributed-client")
+		queryRequest.Trace = true
+		span.Tag("request", convert.BytesToString(logger.Proto(queryRequest)))
+		defer func() {
+			if err != nil {
+				span.Error(err)
+			} else {
+				span.Stop()
+			}
+		}()
+	}
+	ff, err := dctx.Broadcast(defaultQueryTimeout, data.TopicMeasureQuery, bus.NewMessage(bus.MessageID(dctx.TimeRange().Begin.Nanos), queryRequest))
 	if err != nil {
 		return nil, err
 	}
 	var see []sort.Iterator[*comparableDataPoint]
 	for _, f := range ff {
 		if m, getErr := f.Get(); getErr != nil {
-			allErr = multierr.Append(allErr, getErr)
+			err = multierr.Append(err, getErr)
 		} else {
 			d := m.Data()
 			if d == nil {
 				continue
 			}
+			resp := d.(*measurev1.QueryResponse)
+			if span != nil {
+				span.AddSubTrace(resp.Trace)
+			}
 			see = append(see,
-				newSortableElements(d.(*measurev1.QueryResponse).DataPoints,
+				newSortableElements(resp.DataPoints,
 					t.sortByTime, t.sortTagSpec))
 		}
 	}
-	return &sortedMIterator{
+	smi := &sortedMIterator{
 		Iterator: sort.NewItemIter(see, t.desc),
-	}, allErr
+	}
+	smi.init()
+	return smi, err
 }
 
 func (t *distributedPlan) String() string {
@@ -196,8 +219,8 @@ func (t *distributedPlan) Schema() logical.Schema {
 	return t.s
 }
 
-func (t *distributedPlan) Limit(max int) {
-	t.maxDataPointsSize = uint32(max)
+func (t *distributedPlan) Limit(maxVal int) {
+	t.maxDataPointsSize = uint32(maxVal)
 }
 
 var _ sort.Comparable = (*comparableDataPoint)(nil)
@@ -278,8 +301,91 @@ var _ executor.MIterator = (*sortedMIterator)(nil)
 
 type sortedMIterator struct {
 	sort.Iterator[*comparableDataPoint]
+	data        *list.List
+	uniqueData  map[uint64]*measurev1.DataPoint
+	cur         *measurev1.DataPoint
+	initialized bool
+	closed      bool
+}
+
+func (s *sortedMIterator) init() {
+	if s.initialized {
+		return
+	}
+	s.initialized = true
+	if !s.Iterator.Next() {
+		s.closed = true
+		return
+	}
+	s.data = list.New()
+	s.uniqueData = make(map[uint64]*measurev1.DataPoint)
+	s.loadDps()
+}
+
+func (s *sortedMIterator) Next() bool {
+	if s.data == nil {
+		return false
+	}
+	if s.data.Len() == 0 {
+		s.loadDps()
+		if s.data.Len() == 0 {
+			return false
+		}
+	}
+	dp := s.data.Front()
+	s.data.Remove(dp)
+	s.cur = dp.Value.(*measurev1.DataPoint)
+	return true
+}
+
+func (s *sortedMIterator) loadDps() {
+	if s.closed {
+		return
+	}
+	for k := range s.uniqueData {
+		delete(s.uniqueData, k)
+	}
+	first := s.Iterator.Val()
+	s.uniqueData[hashDataPoint(first.DataPoint)] = first.DataPoint
+	for {
+		if !s.Iterator.Next() {
+			s.closed = true
+			break
+		}
+		v := s.Iterator.Val()
+		if bytes.Equal(first.SortedField(), v.SortedField()) {
+			key := hashDataPoint(v.DataPoint)
+			if existed, ok := s.uniqueData[key]; ok {
+				if v.DataPoint.Version > existed.Version {
+					s.uniqueData[key] = v.DataPoint
+				}
+			} else {
+				s.uniqueData[key] = v.DataPoint
+			}
+		} else {
+			break
+		}
+	}
+	for _, v := range s.uniqueData {
+		s.data.PushBack(v)
+	}
 }
 
 func (s *sortedMIterator) Current() []*measurev1.DataPoint {
-	return []*measurev1.DataPoint{s.Val().DataPoint}
+	return []*measurev1.DataPoint{s.cur}
+}
+
+const (
+	offset64 = 14695981039346656037
+	prime64  = 1099511628211
+)
+
+// hashDataPoint calculates the hash value of a data point with fnv64a.
+// https://en.wikipedia.org/wiki/Fowler%E2%80%93Noll%E2%80%93Vo_hash_function
+func hashDataPoint(dp *measurev1.DataPoint) uint64 {
+	h := uint64(offset64)
+	h = (h ^ dp.Sid) * prime64
+	h = (h ^ uint64(dp.Timestamp.Seconds)) * prime64
+	h = (h ^ uint64(dp.Timestamp.Nanos)) * prime64
+	return h
 }
