@@ -28,10 +28,13 @@ import (
 	measurev1 "github.com/apache/skywalking-banyandb/api/proto/banyandb/measure/v1"
 	modelv1 "github.com/apache/skywalking-banyandb/api/proto/banyandb/model/v1"
 	"github.com/apache/skywalking-banyandb/pkg/index"
+	"github.com/apache/skywalking-banyandb/pkg/index/inverted"
 	"github.com/apache/skywalking-banyandb/pkg/logger"
 	pbv1 "github.com/apache/skywalking-banyandb/pkg/pb/v1"
+	"github.com/apache/skywalking-banyandb/pkg/query"
 	"github.com/apache/skywalking-banyandb/pkg/query/executor"
 	"github.com/apache/skywalking-banyandb/pkg/query/logical"
+	"github.com/apache/skywalking-banyandb/pkg/query/model"
 	"github.com/apache/skywalking-banyandb/pkg/timestamp"
 )
 
@@ -48,7 +51,7 @@ type unresolvedIndexScan struct {
 }
 
 func (uis *unresolvedIndexScan) Analyze(s logical.Schema) (logical.Plan, error) {
-	projTags := make([]pbv1.TagProjection, len(uis.projectionTags))
+	projTags := make([]model.TagProjection, len(uis.projectionTags))
 	var projTagsRefs [][]*logical.TagRef
 	if len(uis.projectionTags) > 0 {
 		for i := range uis.projectionTags {
@@ -85,7 +88,7 @@ func (uis *unresolvedIndexScan) Analyze(s logical.Schema) (logical.Plan, error)
 		// fill AnyEntry by default
 		entity[idx] = pbv1.AnyTagValue
 	}
-	filter, entities, err := logical.BuildLocalFilter(uis.criteria, s, entityMap, entity, true)
+	query, entities, _, err := inverted.BuildLocalQuery(uis.criteria, s, entityMap, entity)
 	if err != nil {
 		return nil, err
 	}
@@ -98,7 +101,7 @@ func (uis *unresolvedIndexScan) Analyze(s logical.Schema) (logical.Plan, error)
 		projectionTagsRefs:   projTagsRefs,
 		projectionFieldsRefs: projFieldRefs,
 		metadata:             uis.metadata,
-		filter:               filter,
+		query:                query,
 		entities:             entities,
 		groupByEntity:        uis.groupByEntity,
 		uis:                  uis,
@@ -107,57 +110,53 @@ func (uis *unresolvedIndexScan) Analyze(s logical.Schema) (logical.Plan, error)
 }
 
 var (
-	_ logical.Plan          = (*localIndexScan)(nil)
-	_ logical.Sorter        = (*localIndexScan)(nil)
-	_ logical.VolumeLimiter = (*localIndexScan)(nil)
+	_ logical.Plan   = (*localIndexScan)(nil)
+	_ logical.Sorter = (*localIndexScan)(nil)
 )
 
 type localIndexScan struct {
-	filter               index.Filter
+	query                index.Query
 	schema               logical.Schema
 	uis                  *unresolvedIndexScan
 	order                *logical.OrderBy
 	metadata             *commonv1.Metadata
 	l                    *logger.Logger
 	timeRange            timestamp.TimeRange
-	projectionTags       []pbv1.TagProjection
+	projectionTags       []model.TagProjection
 	projectionTagsRefs   [][]*logical.TagRef
 	projectionFieldsRefs []*logical.FieldRef
 	entities             [][]*modelv1.TagValue
 	projectionFields     []string
-	maxDataPointsSize    int
 	groupByEntity        bool
 }
 
-func (i *localIndexScan) Limit(max int) {
-	i.maxDataPointsSize = max
-}
-
 func (i *localIndexScan) Sort(order *logical.OrderBy) {
 	i.order = order
 }
 
 func (i *localIndexScan) Execute(ctx context.Context) (mit executor.MIterator, err error) {
-	var orderBy *pbv1.OrderBy
-	orderByType := pbv1.OrderByTypeTime
+	var orderBy *model.OrderBy
+	orderByType := model.OrderByTypeTime
 	if i.order != nil {
 		if i.order.Index != nil {
-			orderByType = pbv1.OrderByTypeIndex
+			orderByType = model.OrderByTypeIndex
 		}
-		orderBy = &pbv1.OrderBy{
+		orderBy = &model.OrderBy{
 			Index: i.order.Index,
 			Sort:  i.order.Sort,
 		}
 	}
 	if i.groupByEntity {
-		orderByType = pbv1.OrderByTypeSeries
+		orderByType = model.OrderByTypeSeries
 	}
 	ec := executor.FromMeasureExecutionContext(ctx)
-	result, err := ec.Query(ctx, pbv1.MeasureQueryOptions{
+	ctx, stop := i.startSpan(ctx, query.GetTracer(ctx), orderByType, orderBy)
+	defer stop(err)
+	result, err := ec.Query(ctx, model.MeasureQueryOptions{
 		Name:            i.metadata.GetName(),
 		TimeRange:       &i.timeRange,
 		Entities:        i.entities,
-		Filter:          i.filter,
+		Query:           i.query,
 		OrderByType:     orderByType,
 		Order:           orderBy,
 		TagProjection:   i.projectionTags,
@@ -172,9 +171,9 @@ func (i *localIndexScan) Execute(ctx context.Context) (mit executor.MIterator, e
 }
 
 func (i *localIndexScan) String() string {
-	return fmt.Sprintf("IndexScan: startTime=%d,endTime=%d,Metadata{group=%s,name=%s},conditions=%s; projection=%s; order=%s; limit=%d",
+	return fmt.Sprintf("IndexScan: startTime=%d,endTime=%d,Metadata{group=%s,name=%s},conditions=%s; projection=%s; order=%s;",
 		i.timeRange.Start.Unix(), i.timeRange.End.Unix(), i.metadata.GetGroup(), i.metadata.GetName(),
-		i.filter, logical.FormatTagRefs(", ", i.projectionTagsRefs...), i.order, i.maxDataPointsSize)
+		i.query, logical.FormatTagRefs(", ", i.projectionTagsRefs...), i.order)
 }
 
 func (i *localIndexScan) Children() []logical.Plan {
@@ -203,12 +202,16 @@ func indexScan(startTime, endTime time.Time, metadata *commonv1.Metadata, projec
 }
 
 type resultMIterator struct {
-	result  pbv1.MeasureQueryResult
+	result  model.MeasureQueryResult
+	err     error
 	current []*measurev1.DataPoint
 	i       int
 }
 
 func (ei *resultMIterator) Next() bool {
+	if ei.result == nil {
+		return false
+	}
 	ei.i++
 	if ei.i < len(ei.current) {
 		return true
@@ -218,11 +221,17 @@ func (ei *resultMIterator) Next() bool {
 	if r == nil {
 		return false
 	}
+	if r.Error != nil {
+		ei.err = r.Error
+		return false
+	}
 	ei.current = ei.current[:0]
 	ei.i = 0
 	for i := range r.Timestamps {
 		dp := &measurev1.DataPoint{
 			Timestamp: timestamppb.New(time.Unix(0, r.Timestamps[i])),
+			Sid:       uint64(r.SID),
+			Version:   r.Versions[i],
 		}
 
 		for _, tf := range r.TagFamilies {
@@ -254,6 +263,33 @@ func (ei *resultMIterator) Current() []*measurev1.DataPoint {
 }
 
 func (ei *resultMIterator) Close() error {
-	ei.result.Release()
-	return nil
+	if ei.result != nil {
+		ei.result.Release()
+	}
+	return ei.err
+}
+
+func (i *localIndexScan) startSpan(ctx context.Context, tracer *query.Tracer, orderType model.OrderByType, orderBy *model.OrderBy) (context.Context, func(error)) {
+	if tracer == nil {
+		return ctx, func(error) {}
+	}
+
+	span, ctx := tracer.StartSpan(ctx, "indexScan-%s", i.metadata)
+	sortName := modelv1.Sort_name[int32(orderBy.Sort)]
+	switch orderType {
+	case model.OrderByTypeTime:
+		span.Tag("orderBy", "time "+sortName)
+	case model.OrderByTypeIndex:
+		span.Tag("orderBy", fmt.Sprintf("indexRule:%s", orderBy.Index.Metadata.Name))
+	case model.OrderByTypeSeries:
+		span.Tag("orderBy", "series")
+	}
+	span.Tag("details", i.String())
+
+	return ctx, func(err error) {
+		if err != nil {
+			span.Error(err)
+		}
+		span.Stop()
+	}
 }
