@@ -34,6 +34,7 @@ import (
 
 	"github.com/apache/skywalking-banyandb/api/data"
 	clusterv1 "github.com/apache/skywalking-banyandb/api/proto/banyandb/cluster/v1"
+	databasev1 "github.com/apache/skywalking-banyandb/api/proto/banyandb/database/v1"
 	"github.com/apache/skywalking-banyandb/banyand/metadata"
 	"github.com/apache/skywalking-banyandb/banyand/metadata/schema"
 	"github.com/apache/skywalking-banyandb/banyand/queue"
@@ -52,7 +53,7 @@ type pub struct {
 	metadata   metadata.Repo
 	handler    schema.EventHandler
 	log        *logger.Logger
-	registered map[string]struct{}
+	registered map[string]*databasev1.Node
 	active     map[string]*client
 	evictable  map[string]evictNode
 	closer     *run.Closer
@@ -169,8 +170,8 @@ func (p *pub) publish(timeout time.Duration, topic bus.Topic, messages ...bus.Me
 	return f, err
 }
 
-func (p *pub) Publish(topic bus.Topic, messages ...bus.Message) (bus.Future, error) {
-	return p.publish(5*time.Second, topic, messages...)
+func (p *pub) Publish(_ context.Context, _ bus.Topic, _ ...bus.Message) (bus.Future, error) {
+	panic("should not be called")
 }
 
 // NewBatchPublisher returns a new batch publisher.
@@ -189,7 +190,7 @@ func New(metadata metadata.Repo) queue.Client {
 		metadata:   metadata,
 		active:     make(map[string]*client),
 		evictable:  make(map[string]evictNode),
-		registered: make(map[string]struct{}),
+		registered: make(map[string]*databasev1.Node),
 		closer:     run.NewCloser(1),
 	}
 }
@@ -199,7 +200,7 @@ func (*pub) Name() string {
 }
 
 func (p *pub) PreRun(context.Context) error {
-	p.log = logger.GetLogger("server-queue")
+	p.log = logger.GetLogger("server-queue-pub")
 	p.metadata.RegisterHandler("queue-client", schema.KindNode, p)
 	return nil
 }
@@ -234,7 +235,7 @@ func (bp *batchPublisher) Close() (err error) {
 	return err
 }
 
-func (bp *batchPublisher) Publish(topic bus.Topic, messages ...bus.Message) (bus.Future, error) {
+func (bp *batchPublisher) Publish(ctx context.Context, topic bus.Topic, messages ...bus.Message) (bus.Future, error) {
 	var err error
 	for _, m := range messages {
 		r, errM2R := messageToRequest(topic, m)
@@ -251,6 +252,8 @@ func (bp *batchPublisher) Publish(topic bus.Topic, messages ...bus.Message) (bus
 					}
 				}()
 				select {
+				case <-ctx.Done():
+					return false
 				case <-stream.client.Context().Done():
 					return false
 				default:
@@ -268,6 +271,12 @@ func (bp *batchPublisher) Publish(topic bus.Topic, messages ...bus.Message) (bus
 			continue
 		}
 
+		select {
+		case <-ctx.Done():
+			return nil, ctx.Err()
+		default:
+		}
+
 		bp.pub.mu.RLock()
 		client, ok := bp.pub.active[node]
 		bp.pub.mu.RUnlock()
@@ -275,17 +284,17 @@ func (bp *batchPublisher) Publish(topic bus.Topic, messages ...bus.Message) (bus
 			err = multierr.Append(err, fmt.Errorf("failed to get client for node %s", node))
 			continue
 		}
-		ctx, cancel := context.WithTimeout(context.Background(), bp.timeout)
+		streamCtx, cancel := context.WithTimeout(ctx, bp.timeout)
 		// this assignment is for getting around the go vet lint
 		deferFn := cancel
-		stream, errCreateStream := client.client.Send(ctx)
+		stream, errCreateStream := client.client.Send(streamCtx)
 		if errCreateStream != nil {
 			err = multierr.Append(err, fmt.Errorf("failed to get stream for node %s: %w", node, errCreateStream))
 			continue
 		}
 		bp.streams[node] = writeStream{
 			client:    stream,
-			ctxDoneCh: ctx.Done(),
+			ctxDoneCh: streamCtx.Done(),
 		}
 		bp.f.events = append(bp.f.events, make(chan batchEvent))
 		_ = sendData()
@@ -295,6 +304,11 @@ func (bp *batchPublisher) Publish(topic bus.Topic, messages ...bus.Message) (bus
 				deferFn()
 			}()
 			for {
+				select {
+				case <-ctx.Done():
+					return
+				default:
+				}
 				_, errRecv := s.Recv()
 				if errRecv == nil {
 					continue
