@@ -18,21 +18,27 @@
 package stream
 
 import (
+	"fmt"
 	"sort"
-	"sync"
+
+	"golang.org/x/exp/slices"
 
 	"github.com/apache/skywalking-banyandb/api/common"
+	modelv1 "github.com/apache/skywalking-banyandb/api/proto/banyandb/model/v1"
 	"github.com/apache/skywalking-banyandb/pkg/bytes"
 	"github.com/apache/skywalking-banyandb/pkg/encoding"
 	"github.com/apache/skywalking-banyandb/pkg/fs"
+	"github.com/apache/skywalking-banyandb/pkg/index/posting"
 	"github.com/apache/skywalking-banyandb/pkg/logger"
 	pbv1 "github.com/apache/skywalking-banyandb/pkg/pb/v1"
+	"github.com/apache/skywalking-banyandb/pkg/pool"
+	"github.com/apache/skywalking-banyandb/pkg/query/model"
 	"github.com/apache/skywalking-banyandb/pkg/timestamp"
 )
 
 type block struct {
 	timestamps  []int64
-	elementIDs  []string
+	elementIDs  []uint64
 	tagFamilies []tagFamily
 }
 
@@ -47,7 +53,7 @@ func (b *block) reset() {
 	b.tagFamilies = tff[:0]
 }
 
-func (b *block) mustInitFromElements(timestamps []int64, elementIDs []string, tagFamilies [][]tagValues) {
+func (b *block) mustInitFromElements(timestamps []int64, elementIDs []uint64, tagFamilies [][]tagValues) {
 	b.reset()
 	size := len(timestamps)
 	if size == 0 {
@@ -66,7 +72,7 @@ func (b *block) mustInitFromElements(timestamps []int64, elementIDs []string, ta
 func assertTimestampsSorted(timestamps []int64) {
 	for i := range timestamps {
 		if i > 0 && timestamps[i-1] > timestamps[i] {
-			logger.Panicf("log entries must be sorted by timestamp; got the previous entry with bigger timestamp %d than the current entry with timestamp %d",
+			logger.Panicf("elements must be sorted by timestamp; got the previous element with bigger timestamp %d than the current element with timestamp %d",
 				timestamps[i-1], timestamps[i])
 		}
 	}
@@ -122,8 +128,7 @@ func (b *block) mustWriteTo(sid common.SeriesID, bm *blockMetadata, ww *writers)
 	bm.uncompressedSizeBytes = b.uncompressedSizeBytes()
 	bm.count = uint64(b.Len())
 
-	mustWriteTimestampsTo(&bm.timestamps, b.timestamps, &ww.timestampsWriter)
-	mustWriteElementIDsTo(&bm.elementIDs, b.elementIDs, &ww.elementIDsWriter)
+	mustWriteTimestampsTo(&bm.timestamps, b.timestamps, b.elementIDs, &ww.timestampsWriter)
 
 	for ti := range b.tagFamilies {
 		b.marshalTagFamily(b.tagFamilies[ti], bm, ww)
@@ -229,7 +234,7 @@ func (b *block) unmarshalTagFamilyFromSeqReaders(decoder *encoding.BytesBlockDec
 func (b *block) uncompressedSizeBytes() uint64 {
 	elementsCount := uint64(b.Len())
 
-	n := elementsCount * 8
+	n := elementsCount * (8 + 8) // 8 bytes for timestamp and 8 bytes for elementID
 
 	tff := b.tagFamilies
 	for i := range tff {
@@ -250,8 +255,7 @@ func (b *block) uncompressedSizeBytes() uint64 {
 func (b *block) mustReadFrom(decoder *encoding.BytesBlockDecoder, p *part, bm blockMetadata) {
 	b.reset()
 
-	b.timestamps = mustReadTimestampsFrom(b.timestamps, &bm.timestamps, int(bm.count), p.timestamps)
-	b.elementIDs = mustReadElementIDsFrom(b.elementIDs, &bm.elementIDs, int(bm.count), p.elementIDs)
+	b.timestamps, b.elementIDs = mustReadTimestampsFrom(b.timestamps, b.elementIDs, &bm.timestamps, int(bm.count), p.timestamps)
 
 	_ = b.resizeTagFamilies(len(bm.tagProjection))
 	for i := range bm.tagProjection {
@@ -269,8 +273,7 @@ func (b *block) mustReadFrom(decoder *encoding.BytesBlockDecoder, p *part, bm bl
 func (b *block) mustSeqReadFrom(decoder *encoding.BytesBlockDecoder, seqReaders *seqReaders, bm blockMetadata) {
 	b.reset()
 
-	b.timestamps = mustSeqReadTimestampsFrom(b.timestamps, &bm.timestamps, int(bm.count), &seqReaders.timestamps)
-	b.elementIDs = mustSeqReadElementIDsFrom(b.elementIDs, &bm.elementIDs, int(bm.count), &seqReaders.elementIDs)
+	b.timestamps, b.elementIDs = mustSeqReadTimestampsFrom(b.timestamps, b.elementIDs, &bm.timestamps, int(bm.count), &seqReaders.timestamps)
 
 	_ = b.resizeTagFamilies(len(bm.tagFamilies))
 	keys := make([]string, 0, len(bm.tagFamilies))
@@ -292,71 +295,48 @@ func (b *block) sortTagFamilies() {
 	})
 }
 
-func mustWriteTimestampsTo(tm *timestampsMetadata, timestamps []int64, timestampsWriter *writer) {
+func mustWriteTimestampsTo(tm *timestampsMetadata, timestamps []int64, elementIDs []uint64, timestampsWriter *writer) {
 	tm.reset()
 
 	bb := bigValuePool.Generate()
 	defer bigValuePool.Release(bb)
 	bb.Buf, tm.encodeType, tm.min = encoding.Int64ListToBytes(bb.Buf[:0], timestamps)
-	if len(bb.Buf) > maxTimestampsBlockSize {
-		logger.Panicf("too big block with timestamps: %d bytes; the maximum supported size is %d bytes", len(bb.Buf), maxTimestampsBlockSize)
-	}
 	tm.max = timestamps[len(timestamps)-1]
 	tm.offset = timestampsWriter.bytesWritten
-	tm.size = uint64(len(bb.Buf))
+	tm.elementIDsOffset = uint64(len(bb.Buf))
+	timestampsWriter.MustWrite(bb.Buf)
+	bb.Buf = encoding.VarUint64sToBytes(bb.Buf[:0], elementIDs)
+	tm.size = tm.elementIDsOffset + uint64(len(bb.Buf))
 	timestampsWriter.MustWrite(bb.Buf)
 }
 
-func mustReadTimestampsFrom(dst []int64, tm *timestampsMetadata, count int, reader fs.Reader) []int64 {
+func mustReadTimestampsFrom(timestamps []int64, elementIDs []uint64, tm *timestampsMetadata, count int, reader fs.Reader) ([]int64, []uint64) {
 	bb := bigValuePool.Generate()
 	defer bigValuePool.Release(bb)
 	bb.Buf = bytes.ResizeExact(bb.Buf, int(tm.size))
 	fs.MustReadData(reader, int64(tm.offset), bb.Buf)
+	return mustDecodeTimestampsWithVersions(timestamps, elementIDs, tm, count, reader.Path(), bb.Buf)
+}
+
+func mustDecodeTimestampsWithVersions(timestamps []int64, elementIDs []uint64, tm *timestampsMetadata, count int, path string, src []byte) ([]int64, []uint64) {
+	if tm.size < tm.elementIDsOffset {
+		logger.Panicf("size %d must be greater than elementIDsOffset %d", tm.size, tm.elementIDsOffset)
+	}
 	var err error
-	dst, err = encoding.BytesToInt64List(dst, bb.Buf, tm.encodeType, tm.min, count)
+	timestamps, err = encoding.BytesToInt64List(timestamps, src[:tm.elementIDsOffset], tm.encodeType, tm.min, count)
 	if err != nil {
-		logger.Panicf("%s: cannot unmarshal timestamps: %v", reader.Path(), err)
+		logger.Panicf("%s: cannot unmarshal timestamps: %v", path, err)
 	}
-	return dst
-}
-
-func mustWriteElementIDsTo(em *elementIDsMetadata, elementIDs []string, elementIDsWriter *writer) {
-	em.reset()
-
-	bb := bigValuePool.Generate()
-	defer bigValuePool.Release(bb)
-	elementIDsByteSlice := make([][]byte, len(elementIDs))
-	for i, elementID := range elementIDs {
-		elementIDsByteSlice[i] = []byte(elementID)
-	}
-	bb.Buf = encoding.EncodeBytesBlock(bb.Buf, elementIDsByteSlice)
-	if len(bb.Buf) > maxElementIDsBlockSize {
-		logger.Panicf("too big block with elementIDs: %d bytes; the maximum supported size is %d bytes", len(bb.Buf), maxElementIDsBlockSize)
-	}
-	em.encodeType = encoding.EncodeTypeUnknown
-	em.offset = elementIDsWriter.bytesWritten
-	em.size = uint64(len(bb.Buf))
-	elementIDsWriter.MustWrite(bb.Buf)
-}
-
-func mustReadElementIDsFrom(dst []string, em *elementIDsMetadata, count int, reader fs.Reader) []string {
-	bb := bigValuePool.Generate()
-	defer bigValuePool.Release(bb)
-	bb.Buf = bytes.ResizeExact(bb.Buf, int(em.size))
-	fs.MustReadData(reader, int64(em.offset), bb.Buf)
-	decoder := encoding.BytesBlockDecoder{}
-	var elementIDsByteSlice [][]byte
-	elementIDsByteSlice, err := decoder.Decode(elementIDsByteSlice, bb.Buf, uint64(count))
+	elementIDs = encoding.ExtendListCapacity(elementIDs, count)
+	elementIDs = elementIDs[:count]
+	_, err = encoding.BytesToVarUint64s(elementIDs, src[tm.elementIDsOffset:])
 	if err != nil {
-		logger.Panicf("%s: cannot unmarshal elementIDs: %v", reader.Path(), err)
+		logger.Panicf("%s: cannot unmarshal element ids: %v", path, err)
 	}
-	for _, elementID := range elementIDsByteSlice {
-		dst = append(dst, string(elementID))
-	}
-	return dst
+	return timestamps, elementIDs
 }
 
-func mustSeqReadTimestampsFrom(dst []int64, tm *timestampsMetadata, count int, reader *seqReader) []int64 {
+func mustSeqReadTimestampsFrom(timestamps []int64, elementIDs []uint64, tm *timestampsMetadata, count int, reader *seqReader) ([]int64, []uint64) {
 	if tm.offset != reader.bytesRead {
 		logger.Panicf("offset %d must be equal to bytesRead %d", tm.offset, reader.bytesRead)
 	}
@@ -364,32 +344,7 @@ func mustSeqReadTimestampsFrom(dst []int64, tm *timestampsMetadata, count int, r
 	defer bigValuePool.Release(bb)
 	bb.Buf = bytes.ResizeExact(bb.Buf, int(tm.size))
 	reader.mustReadFull(bb.Buf)
-	var err error
-	dst, err = encoding.BytesToInt64List(dst, bb.Buf, tm.encodeType, tm.min, count)
-	if err != nil {
-		logger.Panicf("%s: cannot unmarshal timestamps: %v", reader.Path(), err)
-	}
-	return dst
-}
-
-func mustSeqReadElementIDsFrom(dst []string, em *elementIDsMetadata, count int, reader *seqReader) []string {
-	if em.offset != reader.bytesRead {
-		logger.Panicf("offset %d must be equal to bytesRead %d", em.offset, reader.bytesRead)
-	}
-	bb := bigValuePool.Generate()
-	defer bigValuePool.Release(bb)
-	bb.Buf = bytes.ResizeExact(bb.Buf, int(em.size))
-	reader.mustReadFull(bb.Buf)
-	decoder := encoding.BytesBlockDecoder{}
-	var elementIDsByteSlice [][]byte
-	elementIDsByteSlice, err := decoder.Decode(elementIDsByteSlice, bb.Buf, uint64(count))
-	if err != nil {
-		logger.Panicf("%s: cannot unmarshal elementIDs: %v", reader.Path(), err)
-	}
-	for _, elementID := range elementIDsByteSlice {
-		dst = append(dst, string(elementID))
-	}
-	return dst
+	return mustDecodeTimestampsWithVersions(timestamps, elementIDs, tm, count, reader.Path(), bb.Buf)
 }
 
 func generateBlock() *block {
@@ -397,7 +352,7 @@ func generateBlock() *block {
 	if v == nil {
 		return &block{}
 	}
-	return v.(*block)
+	return v
 }
 
 func releaseBlock(b *block) {
@@ -405,20 +360,20 @@ func releaseBlock(b *block) {
 	blockPool.Put(b)
 }
 
-var blockPool sync.Pool
+var blockPool = pool.Register[*block]("stream-block")
 
 type blockCursor struct {
-	p                  *part
-	timestamps         []int64
-	expectedTimestamps []int64
-	elementIDs         []string
-	tagFamilies        []tagFamily
-	tagValuesDecoder   encoding.BytesBlockDecoder
-	tagProjection      []pbv1.TagProjection
-	bm                 blockMetadata
-	idx                int
-	minTimestamp       int64
-	maxTimestamp       int64
+	p                *part
+	timestamps       []int64
+	elementFilter    posting.List
+	elementIDs       []uint64
+	tagFamilies      []tagFamily
+	tagValuesDecoder encoding.BytesBlockDecoder
+	tagProjection    []model.TagProjection
+	bm               blockMetadata
+	idx              int
+	minTimestamp     int64
+	maxTimestamp     int64
 }
 
 func (bc *blockCursor) reset() {
@@ -446,67 +401,70 @@ func (bc *blockCursor) init(p *part, bm *blockMetadata, opts queryOptions) {
 	bc.minTimestamp = opts.minTimestamp
 	bc.maxTimestamp = opts.maxTimestamp
 	bc.tagProjection = opts.TagProjection
-	if opts.elementRefMap != nil {
-		seriesID := bc.bm.seriesID
-		bc.expectedTimestamps = opts.elementRefMap[seriesID]
-	}
+	bc.elementFilter = opts.elementFilter
 }
 
-func (bc *blockCursor) copyAllTo(r *pbv1.StreamResult, desc bool) {
-	var idx, offset int
-	if desc {
-		idx = 0
-		offset = bc.idx + 1
-	} else {
-		idx = bc.idx
-		offset = len(bc.timestamps)
+func (bc *blockCursor) copyAllTo(r *model.StreamResult, desc bool) {
+	start, end := 0, bc.idx+1
+	if !desc {
+		start, end = bc.idx, len(bc.timestamps)
 	}
-	if offset <= idx {
+	if end <= start {
 		return
 	}
-	r.SID = bc.bm.seriesID
-	r.Timestamps = append(r.Timestamps, bc.timestamps[idx:offset]...)
-	r.ElementIDs = append(r.ElementIDs, bc.elementIDs[idx:offset]...)
+
+	r.Timestamps = append(r.Timestamps, bc.timestamps[start:end]...)
+	r.ElementIDs = append(r.ElementIDs, bc.elementIDs[start:end]...)
+	requiredCapacity := end - start
+	r.SIDs = append(r.SIDs, make([]common.SeriesID, requiredCapacity)...)
+	for i := range r.SIDs[len(r.SIDs)-requiredCapacity:] {
+		r.SIDs[len(r.SIDs)-requiredCapacity+i] = bc.bm.seriesID
+	}
+
+	if desc {
+		slices.Reverse(r.Timestamps)
+		slices.Reverse(r.ElementIDs)
+	}
+
 	if len(r.TagFamilies) != len(bc.tagProjection) {
-		for _, tp := range bc.tagProjection {
-			tf := pbv1.TagFamily{
-				Name: tp.Family,
+		r.TagFamilies = make([]model.TagFamily, len(bc.tagProjection))
+		for i, tp := range bc.tagProjection {
+			r.TagFamilies[i] = model.TagFamily{Name: tp.Family, Tags: make([]model.Tag, len(tp.Names))}
+			for j, n := range tp.Names {
+				r.TagFamilies[i].Tags[j] = model.Tag{Name: n}
 			}
-			for _, n := range tp.Names {
-				t := pbv1.Tag{
-					Name: n,
-				}
-				tf.Tags = append(tf.Tags, t)
-			}
-			r.TagFamilies = append(r.TagFamilies, tf)
 		}
 	}
+
 	for i, cf := range bc.tagFamilies {
-		for i2, c := range cf.tags {
-			if c.values != nil {
-				for _, v := range c.values[idx:offset] {
-					r.TagFamilies[i].Tags[i2].Values = append(r.TagFamilies[i].Tags[i2].Values, mustDecodeTagValue(c.valueType, v))
-				}
-			} else {
-				for j := idx; j < offset; j++ {
-					r.TagFamilies[i].Tags[i2].Values = append(r.TagFamilies[i].Tags[i2].Values, pbv1.NullTagValue)
+		for j, c := range cf.tags {
+			values := make([]*modelv1.TagValue, end-start)
+			for k := start; k < end; k++ {
+				if c.values != nil {
+					values[k-start] = mustDecodeTagValue(c.valueType, c.values[k])
+				} else {
+					values[k-start] = pbv1.NullTagValue
 				}
 			}
+			if desc {
+				slices.Reverse(values)
+			}
+			r.TagFamilies[i].Tags[j].Values = append(r.TagFamilies[i].Tags[j].Values, values...)
 		}
 	}
 }
 
-func (bc *blockCursor) copyTo(r *pbv1.StreamResult) {
-	r.SID = bc.bm.seriesID
+func (bc *blockCursor) copyTo(r *model.StreamResult) {
 	r.Timestamps = append(r.Timestamps, bc.timestamps[bc.idx])
 	r.ElementIDs = append(r.ElementIDs, bc.elementIDs[bc.idx])
+	r.SIDs = append(r.SIDs, bc.bm.seriesID)
 	if len(r.TagFamilies) != len(bc.tagProjection) {
 		for _, tp := range bc.tagProjection {
-			tf := pbv1.TagFamily{
+			tf := model.TagFamily{
 				Name: tp.Family,
 			}
 			for _, n := range tp.Names {
-				t := pbv1.Tag{
+				t := model.Tag{
 					Name: n,
 				}
 				tf.Tags = append(tf.Tags, t)
@@ -550,15 +508,13 @@ func (bc *blockCursor) loadData(tmpBlock *block) bool {
 
 	idxList := make([]int, 0)
 	var start, end int
-	if bc.expectedTimestamps != nil {
-		for _, ts := range bc.expectedTimestamps {
-			idx := timestamp.Find(tmpBlock.timestamps, ts)
-			if idx == -1 {
-				continue
+	if bc.elementFilter != nil {
+		for i := range tmpBlock.elementIDs {
+			if bc.elementFilter.Contains(tmpBlock.elementIDs[i]) {
+				idxList = append(idxList, i)
+				bc.timestamps = append(bc.timestamps, tmpBlock.timestamps[i])
+				bc.elementIDs = append(bc.elementIDs, tmpBlock.elementIDs[i])
 			}
-			idxList = append(idxList, idx)
-			bc.timestamps = append(bc.timestamps, tmpBlock.timestamps[idx])
-			bc.elementIDs = append(bc.elementIDs, tmpBlock.elementIDs[idx])
 		}
 		if len(bc.timestamps) == 0 {
 			return false
@@ -588,7 +544,7 @@ func (bc *blockCursor) loadData(tmpBlock *block) bool {
 					logger.Panicf("unexpected number of values for tags %q: got %d; want %d",
 						tmpBlock.tagFamilies[i].tags[blockIndex].name, len(tmpBlock.tagFamilies[i].tags[blockIndex].values), len(tmpBlock.timestamps))
 				}
-				if bc.expectedTimestamps != nil {
+				if len(idxList) > 0 {
 					for _, idx := range idxList {
 						t.values = append(t.values, tmpBlock.tagFamilies[i].tags[blockIndex].values[idx])
 					}
@@ -604,14 +560,14 @@ func (bc *blockCursor) loadData(tmpBlock *block) bool {
 	return true
 }
 
-var blockCursorPool sync.Pool
+var blockCursorPool = pool.Register[*blockCursor]("stream-blockCursor")
 
 func generateBlockCursor() *blockCursor {
 	v := blockCursorPool.Get()
 	if v == nil {
 		return &blockCursor{}
 	}
-	return v.(*blockCursor)
+	return v
 }
 
 func releaseBlockCursor(bc *blockCursor) {
@@ -648,39 +604,20 @@ func (bi *blockPointer) appendAll(b *blockPointer) {
 	bi.append(b, len(b.timestamps))
 }
 
+var log = logger.GetLogger("stream").Named("block")
+
 func (bi *blockPointer) append(b *blockPointer, offset int) {
 	if offset <= b.idx {
 		return
 	}
 	if len(bi.tagFamilies) == 0 && len(b.tagFamilies) > 0 {
-		for _, tf := range b.tagFamilies {
-			tFamily := tagFamily{name: tf.name}
-			for _, c := range tf.tags {
-				col := tag{name: c.name, valueType: c.valueType}
-				assertIdxAndOffset(col.name, len(c.values), b.idx, offset)
-				col.values = append(col.values, c.values[b.idx:offset]...)
-				tFamily.tags = append(tFamily.tags, col)
-			}
-			bi.tagFamilies = append(bi.tagFamilies, tFamily)
-		}
+		fullTagAppend(bi, b, offset)
 	} else {
-		if len(bi.tagFamilies) != len(b.tagFamilies) {
-			logger.Panicf("unexpected number of tag families: got %d; want %d", len(bi.tagFamilies), len(b.tagFamilies))
-		}
-		for i := range bi.tagFamilies {
-			if bi.tagFamilies[i].name != b.tagFamilies[i].name {
-				logger.Panicf("unexpected tag family name: got %q; want %q", bi.tagFamilies[i].name, b.tagFamilies[i].name)
-			}
-			if len(bi.tagFamilies[i].tags) != len(b.tagFamilies[i].tags) {
-				logger.Panicf("unexpected number of tags for tag family %q: got %d; want %d", bi.tagFamilies[i].name, len(bi.tagFamilies[i].tags), len(b.tagFamilies[i].tags))
-			}
-			for j := range bi.tagFamilies[i].tags {
-				if bi.tagFamilies[i].tags[j].name != b.tagFamilies[i].tags[j].name {
-					logger.Panicf("unexpected tag name for tag family %q: got %q; want %q", bi.tagFamilies[i].name, bi.tagFamilies[i].tags[j].name, b.tagFamilies[i].tags[j].name)
-				}
-				assertIdxAndOffset(b.tagFamilies[i].tags[j].name, len(b.tagFamilies[i].tags[j].values), b.idx, offset)
-				bi.tagFamilies[i].tags[j].values = append(bi.tagFamilies[i].tags[j].values, b.tagFamilies[i].tags[j].values[b.idx:offset]...)
+		if err := fastTagAppend(bi, b, offset); err != nil {
+			if log.Debug().Enabled() {
+				log.Debug().Msgf("fastTagMerge failed: %v; falling back to fullTagMerge", err)
 			}
+			fullTagAppend(bi, b, offset)
 		}
 	}
 
@@ -689,6 +626,113 @@ func (bi *blockPointer) append(b *blockPointer, offset int) {
 	bi.elementIDs = append(bi.elementIDs, b.elementIDs[b.idx:offset]...)
 }
 
+func fastTagAppend(bi, b *blockPointer, offset int) error {
+	if len(bi.tagFamilies) != len(b.tagFamilies) {
+		return fmt.Errorf("unexpected number of tag families: got %d; want %d", len(b.tagFamilies), len(bi.tagFamilies))
+	}
+	for i := range bi.tagFamilies {
+		if bi.tagFamilies[i].name != b.tagFamilies[i].name {
+			return fmt.Errorf("unexpected tag family name: got %q; want %q", b.tagFamilies[i].name, bi.tagFamilies[i].name)
+		}
+		if len(bi.tagFamilies[i].tags) != len(b.tagFamilies[i].tags) {
+			return fmt.Errorf("unexpected number of tags for tag family %q: got %d; want %d",
+				bi.tagFamilies[i].name, len(b.tagFamilies[i].tags), len(bi.tagFamilies[i].tags))
+		}
+		for j := range bi.tagFamilies[i].tags {
+			if bi.tagFamilies[i].tags[j].name != b.tagFamilies[i].tags[j].name {
+				return fmt.Errorf("unexpected tag name for tag family %q: got %q; want %q",
+					bi.tagFamilies[i].name, b.tagFamilies[i].tags[j].name, bi.tagFamilies[i].tags[j].name)
+			}
+			assertIdxAndOffset(b.tagFamilies[i].tags[j].name, len(b.tagFamilies[i].tags[j].values), b.idx, offset)
+			bi.tagFamilies[i].tags[j].values = append(bi.tagFamilies[i].tags[j].values, b.tagFamilies[i].tags[j].values[b.idx:offset]...)
+		}
+	}
+	return nil
+}
+
+func fullTagAppend(bi, b *blockPointer, offset int) {
+	existDataSize := len(bi.timestamps)
+	appendTagFamilies := func(tf tagFamily) {
+		tfv := tagFamily{name: tf.name}
+		for i := range tf.tags {
+			assertIdxAndOffset(tf.tags[i].name, len(tf.tags[i].values), b.idx, offset)
+			col := tag{name: tf.tags[i].name, valueType: tf.tags[i].valueType}
+			for j := 0; j < existDataSize; j++ {
+				col.values = append(col.values, nil)
+			}
+			col.values = append(col.values, tf.tags[i].values[b.idx:offset]...)
+			tfv.tags = append(tfv.tags, col)
+		}
+		bi.tagFamilies = append(bi.tagFamilies, tfv)
+	}
+	if len(bi.tagFamilies) == 0 {
+		for _, tf := range b.tagFamilies {
+			appendTagFamilies(tf)
+		}
+		return
+	}
+
+	tagFamilyMap := make(map[string]*tagFamily)
+	for i := range bi.tagFamilies {
+		tagFamilyMap[bi.tagFamilies[i].name] = &bi.tagFamilies[i]
+	}
+
+	for _, tf := range b.tagFamilies {
+		if existingTagFamily, exists := tagFamilyMap[tf.name]; exists {
+			columnMap := make(map[string]*tag)
+			for i := range existingTagFamily.tags {
+				columnMap[existingTagFamily.tags[i].name] = &existingTagFamily.tags[i]
+			}
+
+			for _, c := range tf.tags {
+				if existingColumn, exists := columnMap[c.name]; exists {
+					assertIdxAndOffset(c.name, len(c.values), b.idx, offset)
+					existingColumn.values = append(existingColumn.values, c.values[b.idx:offset]...)
+				} else {
+					assertIdxAndOffset(c.name, len(c.values), b.idx, offset)
+					col := tag{name: c.name, valueType: c.valueType}
+					for j := 0; j < existDataSize; j++ {
+						col.values = append(col.values, nil)
+					}
+					col.values = append(col.values, c.values[b.idx:offset]...)
+					existingTagFamily.tags = append(existingTagFamily.tags, col)
+				}
+			}
+		} else {
+			appendTagFamilies(tf)
+		}
+	}
+	for k := range tagFamilyMap {
+		delete(tagFamilyMap, k)
+	}
+	for i := range b.tagFamilies {
+		tagFamilyMap[b.tagFamilies[i].name] = &b.tagFamilies[i]
+	}
+	emptySize := offset - b.idx
+	for _, tf := range bi.tagFamilies {
+		if _, exists := tagFamilyMap[tf.name]; !exists {
+			for i := range tf.tags {
+				for j := 0; j < emptySize; j++ {
+					tf.tags[i].values = append(tf.tags[i].values, nil)
+				}
+			}
+		} else {
+			existingTagFamily := tagFamilyMap[tf.name]
+			columnMap := make(map[string]*tag)
+			for i := range existingTagFamily.tags {
+				columnMap[existingTagFamily.tags[i].name] = &existingTagFamily.tags[i]
+			}
+			for i := range tf.tags {
+				if _, exists := columnMap[tf.tags[i].name]; !exists {
+					for j := 0; j < emptySize; j++ {
+						tf.tags[i].values = append(tf.tags[i].values, nil)
+					}
+				}
+			}
+		}
+	}
+}
+
 func assertIdxAndOffset(name string, length int, idx int, offset int) {
 	if idx >= offset {
 		logger.Panicf("%q idx %d must be less than offset %d", name, idx, offset)
@@ -713,7 +757,7 @@ func generateBlockPointer() *blockPointer {
 	if v == nil {
 		return &blockPointer{}
 	}
-	return v.(*blockPointer)
+	return v
 }
 
 func releaseBlockPointer(bi *blockPointer) {
@@ -721,4 +765,4 @@ func releaseBlockPointer(bi *blockPointer) {
 	blockPointerPool.Put(bi)
 }
 
-var blockPointerPool sync.Pool
+var blockPointerPool = pool.Register[*blockPointer]("stream-blockPointer")
