@@ -36,8 +36,10 @@ import (
 	streamv1 "github.com/apache/skywalking-banyandb/api/proto/banyandb/stream/v1"
 	"github.com/apache/skywalking-banyandb/pkg/grpchelper"
 	"github.com/apache/skywalking-banyandb/pkg/logger"
+	"github.com/apache/skywalking-banyandb/pkg/pool"
 	"github.com/apache/skywalking-banyandb/pkg/test"
 	"github.com/apache/skywalking-banyandb/pkg/test/flags"
+	"github.com/apache/skywalking-banyandb/pkg/test/gmatcher"
 	"github.com/apache/skywalking-banyandb/pkg/test/helpers"
 	"github.com/apache/skywalking-banyandb/pkg/test/setup"
 	cases_stream_data "github.com/apache/skywalking-banyandb/test/cases/stream/data"
@@ -129,7 +131,7 @@ var _ = Describe("Load Test Suit", func() {
 		allQueryLatencyData := make([]float64, 0)
 		for i := 0; i < minutes; i++ {
 			GinkgoWriter.Printf("writing data at %s\n", now)
-			cases_stream_data.Write(connection, "data.json", now, interval)
+			cases_stream_data.Write(connection, "sw", now, interval)
 			if now.Sub(lastQueryTime) > queryInterval {
 				latency := queryFn(now, time.Hour)
 				latest1HourQueryLatencyData = append(latest1HourQueryLatencyData, float64(latency.Milliseconds()))
@@ -157,17 +159,18 @@ var _ = Describe("Load Test Suit", func() {
 		}
 		deferFunc()
 		Eventually(gleak.Goroutines, flags.EventuallyTimeout).ShouldNot(gleak.HaveLeaked(goods))
+		Eventually(pool.AllRefsCount, flags.EventuallyTimeout).Should(gmatcher.HaveZeroRef())
 	})
 })
 
 func analysis(name string, data []float64) {
-	min, _ := stats.Min(data)
-	max, _ := stats.Max(data)
+	minVal, _ := stats.Min(data)
+	maxVal, _ := stats.Max(data)
 	mean, _ := stats.Mean(data)
 	median, _ := stats.Median(data)
 	p90, _ := stats.Percentile(data, 90)
 	p95, _ := stats.Percentile(data, 95)
 	p98, _ := stats.Percentile(data, 98)
 	p99, _ := stats.Percentile(data, 99)
-	logger.Infof("%s: min: %f, max: %f, mean: %f, median: %f, p90: %f, p95: %f, p98: %f, p99: %f\n", name, min, max, mean, median, p90, p95, p98, p99)
+	logger.Infof("%s: min: %f, max: %f, mean: %f, median: %f, p90: %f, p95: %f, p98: %f, p99: %f\n", name, minVal, maxVal, mean, median, p90, p95, p98, p99)
 }
